import gzip
import os
from collections import defaultdict

import scipy.io
import scipy.sparse
import pandas as pd
import numpy as np

from celescope.tools.__init__ import (BARCODE_FILE_NAME, FEATURE_FILE_NAME, MATRIX_FILE_NAME)
from celescope.tools import utils

ROW = 'geneID'
COLUMN = 'Barcode'

  
class Features:
    def __init__(self, gene_id: list, gene_name=None, gene_type=None):
        """
        Args:
            gene_id: list of gene id
            gene_name: list of gene name
            type: ype of features, e.g. [gene, protein]
        """
        self.gene_id = list(gene_id)
        if not gene_name:
            self.gene_name = self.gene_id
        else:
            self.gene_name = list(gene_name)
        self.gene_type = gene_type 

    @classmethod
    def from_tsv(cls, tsv_file):
        if (not tsv_file) or (not os.path.exists(tsv_file)):
            raise FileNotFoundError(f'ERROR: {tsv_file} does not exist. \nSOLUTION: Rename and gzip the features file to {FEATURE_FILE_NAME}\n')
        df = pd.read_csv(tsv_file, sep='\t', on_bad_lines='skip', names=['gene_id', 'gene_name', 'type'], dtype=str)
        gene_id = df['gene_id'].tolist()
        gene_name = df['gene_name'].tolist()
        # avoid adding extra \t to genes.tsv when all the gene_type are Nan
        # if gene_type is None and add to dataframe, will cause Seurat::Read error: Error in FUN(X[[i]], ...) : # # subscript out of bounds
        if df['type'].isnull().sum() == len(df['type']):
            gene_type = None
        else:
            gene_type = df['type'].tolist()
        return cls(gene_id, gene_name, gene_type)

    def to_tsv(self, tsv_file):
        """
        if gene_type is None and add to dataframe, will cause Seurat::Read10X error: Error in FUN(X[[i]], ...) : subscript out of bounds
        """
        if self.gene_type:
            df = pd.DataFrame({'gene_id': self.gene_id, 'gene_name': self.gene_name, 'gene_type': self.gene_type})
        else:
            df = pd.DataFrame({'gene_id': self.gene_id, 'gene_name': self.gene_name})
        df.to_csv(tsv_file, sep='\t', index=False, header=False)


class CountMatrix:
    def __init__(self, features: Features, barcodes: list, matrix):
        """
        Args:
            features: Features object
            barcodes: list of barcodes
            matrix: scipy.sparse.coo_matrix
        """
        self.__features = features
        self.__barcodes = barcodes
        self.__matrix = matrix
        self.shape = matrix.shape
    
    @staticmethod
    @utils.add_log
    def read_barcodes(matrix_dir):
        """
            returns: barcodes set
        """
        barcode_file = utils.get_matrix_file_path(matrix_dir, BARCODE_FILE_NAME)
        barcodes, _ = utils.read_one_col(barcode_file)
        return set(barcodes)

    @classmethod
    @utils.add_log
    def from_matrix_dir(cls, matrix_dir):
        if not os.path.exists(matrix_dir):
            raise FileNotFoundError(f'{matrix_dir} does not exist')
        features_tsv = utils.get_matrix_file_path(matrix_dir, FEATURE_FILE_NAME)
        features = Features.from_tsv(tsv_file=features_tsv)
        barcode_file = utils.get_matrix_file_path(matrix_dir, BARCODE_FILE_NAME)
        barcodes, _ = utils.read_one_col(barcode_file)
        matrix_path = utils.get_matrix_file_path(matrix_dir, MATRIX_FILE_NAME)
        matrix= scipy.io.mmread(matrix_path)

        return cls(features, barcodes, matrix)

    @utils.add_log
    def to_matrix_dir(self, matrix_dir):
        utils.check_mkdir(dir_name=matrix_dir)
        self.__features.to_tsv(f'{matrix_dir}/{FEATURE_FILE_NAME}')
        pd.Series(self.__barcodes).to_csv(f'{matrix_dir}/{BARCODE_FILE_NAME}', index=False, sep='\t', header=False)
        matrix_path = f'{matrix_dir}/{MATRIX_FILE_NAME}'
        with gzip.open(matrix_path, 'wb') as f:
            scipy.io.mmwrite(f, self.__matrix)

    @classmethod
    def from_dataframe(cls, df, features: Features, barcodes=None, value="UMI"):
        """
        Use all gene_id from features even if it is not in df
        Args:
            df: count details dataframe with columns UMI and multi-index (column, row).
            features: Features
            barcodes: only keep barcode in barcodes
            value: value name in df, UMI
            
        """
        if not barcodes:
            barcodes = df.index.levels[0].tolist()
        
        feature_index_dict = {}
        for index, gene_id in enumerate(features.gene_id):
            feature_index_dict[gene_id] = index

        barcode_index_dict = {}
        for index, barcode in enumerate(barcodes):
            barcode_index_dict[barcode] = index

        # use all barcodes
        barcode_codes = [barcode_index_dict[barcode] for barcode in df.index.get_level_values(level=0)]
        # use all gene_id from features even if it is not in df
        gene_id_codes = [feature_index_dict[gene_id] for gene_id in df.index.get_level_values(level=1)]
        mtx = scipy.sparse.coo_matrix((df[value], (gene_id_codes, barcode_codes)), 
            shape=(len(features.gene_id), len(barcodes)))

        return cls(features, barcodes, mtx)

    def __str__(self):
        n_row, n_col = self.shape[0], self.shape[1]
        return f"CountMatrix object\n {n_row} x {n_col} coo_matrix"


    def __repr__(self):
        return self.__str__()

    def concat_by_barcodes(self, other):
        if other.get_barcodes() != self.get_barcodes():
            raise ValueError('barcodes are not the same')

        if inter := set(self.get_features().gene_id).intersection(set(other.get_features().gene_id)):
            raise ValueError(f'deuplicated gene_id: {inter}')

        gene_id = self.get_features().gene_id + other.get_features().gene_id
        gene_name = self.get_features().gene_name + other.get_features().gene_name
        gene_type = None
        if self.get_features().gene_type and other.get_features().gene_type:
            gene_type = self.get_features().gene_type + other.get_features().gene_type
        features = Features(gene_id, gene_name, gene_type)

        matrix = scipy.sparse.vstack([self.get_matrix(), other.get_matrix()])

        return CountMatrix(features, self.__barcodes, matrix)

    @utils.add_log
    def slice_matrix(self, slice_barcodes_indices):
        """
        Args:
            slice_barcodes_indices: list of barcode indices
        Returns:
            CountMatrix object
        """
        mtx_csc = self.__matrix.tocsc()
        slice_barcodes_indices.sort()
        sliced_mtx = mtx_csc[:, slice_barcodes_indices]
        barcodes = [self.__barcodes[i] for i in slice_barcodes_indices]
        return CountMatrix(self.__features, barcodes, sliced_mtx)
    
    @utils.add_log
    def slice_matrix_bc(self, bcs):
        """
        Args:
            bcs: cell barcodes
        Returns:
            CountMatrix object
        """
        barcodes_indices = [self.__barcodes.index(barcode) for barcode in bcs]
        barcodes_indices.sort()
        return self.slice_matrix(barcodes_indices)

    @utils.add_log
    def get_genes_fraction(self, gene_list):
        """
        Returns:
            numpy 2d fraction of gene_names in gene_list
        """
        gene_indices = [self.__features.gene_name.index(gene) for gene in gene_list]
        mtx = self.__matrix.tocsc()
        total = mtx.sum(axis=0)
        gene = mtx[gene_indices,:].sum(axis=0)
        f = gene / total

        return f

    def get_bc_geneNum(self):
        """
        Returns {bc: geneNum}, total_genes
        """
        gene_index, bc_index = self.__matrix.nonzero()
        total_genes = len(set(gene_index))
        bc_gene = defaultdict(set)
        for gene, bc in zip(gene_index, bc_index):
            bc_gene[bc].add(gene)
        return {bc: len(bc_gene[bc]) for bc in bc_gene}, total_genes

    def get_barcodes(self):
        return self.__barcodes

    def get_features(self):
        return self.__features
    
    def get_matrix(self):
        return self.__matrix

    @classmethod
    def dataframe_to_matrix(cls, df, features: Features, barcodes=None, barcode_column='Barcode', feature_column='geneID'):
        """Convert a counts dataframe to a sparse counts matrix.
        """
        barcode_indices = {barcode: i for i, barcode in enumerate(barcodes)}
        feature_indices = {feature: i for i, feature in enumerate(features.gene_id)}

        matrix = scipy.sparse.lil_matrix((len(barcodes), len(features.gene_id)), dtype=np.float32)
        for (barcode, feature), count in df.groupby([barcode_column, feature_column], sort=False,
                                                        observed=True).size().items():
            matrix[barcode_indices[barcode], feature_indices[feature]] = count

        return matrix.tocsr()



#!/usr/bin/env python3 

import sys
import argparse

parser = argparse.ArgumentParser(description="Filter TRUST4 barcode report file for noise like diffused mRNAs. Output to stdout")
parser.add_argument('-b', help="barcode_report file", required=True, dest="barcode_report")
parser.add_argument('-a', help="annotation file", dest="annot")
parser.add_argument('--highAbund', help="The minimum abundance to be regarded as potential source of diffusion", default=50.0, dest="highAbund")
parser.add_argument('--diffuseFrac', help="The maximum fraction of the diffusion source abundance to be regarded as noise", default=0.02, dest="diffuseFrac")

args = parser.parse_args()

barcodeReport = args.barcode_report
barcodeInfo = {}
highAbundCdr3Info = {}
highAbund = float(args.highAbund)
diffuseFrac = float(args.diffuseFrac)
assembly = {}

fp = open(barcodeReport)
for line in fp:
	if (line[0] == '#'):
		continue
	cols = line.rstrip().split()
	chain1Cols = cols[2].split(',')
	chain2Cols = cols[3].split(',')
#AGAGTGGTCTATCCTA-1	abT	TRBV6-6*01,TRBD2*01,TRBJ2-5*01,*,TGTGCCAGTCTACTTGGGGGGACCCAGTACTTC,CASLLGGTQYF,3.61,AGAGTGGTCTATCCTA-1_48281,100.00,0	TRAV10*01,*,TRAJ34*01,*,TGTGTGGTGAGCGCCCGCACCGACAAGCTCATCTTT,CVVSARTDKLIF,1.00,AGAGTGGTCTATCCTA-1_65563,100.00,0	*	*
	if (len(chain1Cols) > 1 and float(chain1Cols[6]) >= highAbund):
		cdr3 = chain1Cols[4]
		abund = float(chain1Cols[6])
		if (cdr3 not in highAbundCdr3Info):
			highAbundCdr3Info[cdr3] = {}
		highAbundCdr3Info[cdr3][cols[0]] = [0, abund] 
	if (len(chain1Cols) > 1):
		assembly[chain1Cols[7]] = chain1Cols[4]
		chain1Cols[6] = float(chain1Cols[6])

	if (len(chain2Cols) > 1 and float(chain2Cols[6]) >= highAbund):
		cdr3 = chain2Cols[4]
		abund = float(chain2Cols[6])
		if (cdr3 not in highAbundCdr3Info):
			highAbundCdr3Info[cdr3] = {}
		highAbundCdr3Info[cdr3][cols[0]] = [1, abund]
	if (len(chain2Cols) > 1):
		assembly[chain2Cols[7]] = chain2Cols[4]
		chain2Cols[6]	 = float(chain2Cols[6])
	barcodeInfo[cols[0]] = {}
	barcodeInfo[cols[0]]["chain1"] = chain1Cols[:]
	barcodeInfo[cols[0]]["chain2"] = chain2Cols[:]
fp.close()

if (args.annot is not None):
	fp = open(args.annot)
	for line in fp:
		header = line.rstrip()
		seq = fp.readline().rstrip() ;
		assemblyId = header.split()[0][1:]
		if (assemblyId in assembly):
			assembly[assemblyId] = seq
	fp.close()

fp = open(barcodeReport)
for line in fp:
	if (line[0] == '#'):
		print(line.rstrip())
		continue

	cols = line.rstrip().split()
	chain1Cols = cols[2].split(',')
	chain2Cols = cols[3].split(',')

	testAgainst = {}
	if (len(chain1Cols) > 1 and float(chain1Cols[6]) < highAbund and chain1Cols[4] in highAbundCdr3Info):
		cdr3 = chain1Cols[4]
		for bc in highAbundCdr3Info[cdr3].keys():
			if (highAbundCdr3Info[cdr3][bc][0] == 0 and highAbundCdr3Info[cdr3][bc][1] * diffuseFrac > float(chain1Cols[6])):
				testAgainst[bc] = 1
	if (len(chain2Cols) > 1 and float(chain2Cols[6]) < highAbund and chain2Cols[4] in highAbundCdr3Info):
		cdr3 = chain2Cols[4]
		for bc in highAbundCdr3Info[cdr3]:
			if (highAbundCdr3Info[cdr3][bc][0] == 1 and highAbundCdr3Info[cdr3][bc][1] * diffuseFrac > float(chain2Cols[6])):
				testAgainst[bc] = 1
	flagFilter = 0
	for bc in testAgainst:
		testChain1 = barcodeInfo[bc]["chain1"]
		testChain2 = barcodeInfo[bc]["chain2"]
		
		for i in [1, 2]:
			c1 = chain1Cols
			c2 = testChain1
			if (i == 2):
				c1 = chain2Cols
				c2 = testChain2
			if (len(c1) > 1 and len(c2) > 1):
				if (c2[6] * diffuseFrac > float(c1[6]) and assembly[c1[7]] in assembly[c2[7]]):
					flagFilter |= i
			elif (len(c2) > 1):
				flagFilter |= i
	if (flagFilter != 3):
 		print(line.rstrip())
fp.close()







import configparser
import subprocess
import sys
import os
import math

from celescope.tools.__init__ import GENOME_CONFIG
from celescope.__init__ import HELP_DICT, __VERSION__
from celescope.tools import utils

class MakeRef:
    def __init__(self, genome_type, args):
        self.config_file = GENOME_CONFIG

        self.args = args
        self.thread = args.thread
        self.dry_run = args.dry_run

        self.files = {}
        self.meta = {}
        self.meta['genome_name'] = args.genome_name
        self.meta['genome_type'] = genome_type
        self.meta['celescope_version'] = __VERSION__

    def write_config(self):
        config = configparser.ConfigParser()
        config.optionxform = str
        config['files'] = self.files
        config['meta'] = self.meta

        with open(GENOME_CONFIG, 'w') as config_handle:
            config.write(config_handle)

    @staticmethod
    def get_config(genomeDir):
        '''
        add genomeDir prefix to files
        '''
        config_file = f'{genomeDir}/{GENOME_CONFIG}'
        if not os.path.exists(config_file):
            sys.exit(
                f'Error: {config_file} not found.\n'
                'Solution: Use the mkref command of CeleScope to generate the genome.\n'
            )
        config = configparser.ConfigParser()
        config.optionxform = str
        config.read(config_file)
        if 'meta' not in config:
            sys.exit(
                'Error: CeleScope version >= 2.0.0 have upgraded the STAR version, so the old genome can no longer be used.\n'
                'Solution: Use the mkref command of CeleScope to regenerate the genome.\n'
                f'Genome path: {genomeDir}\n'
            )
        for k,v in dict(config['files']).items():
            if v and v != 'None':
                config['files'][k] = f'{genomeDir}/{v}'
            else:
                config['files'][k] = ""
        return config

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.write_config()

    @staticmethod
    def opts(parser, sub_program):
        if sub_program:
            parser.add_argument("--thread", help="Default=6. Threads to use.", default=6)
            parser.add_argument("--genome_name", help="Required, genome name. ", required=True)
            parser.add_argument("--dry_run", help="Only write config file and exit.", action='store_true')


class MakeRef_STAR(MakeRef):
    def __init__(self, genome_type, args):
        super().__init__(genome_type, args)
        self.fasta = args.fasta
        self.STAR_param = args.STAR_param
        self.files['fasta'] = args.fasta
        self.meta['STAR_param'] = args.STAR_param

    @staticmethod
    def opts(parser, sub_program):
        MakeRef.opts(parser, sub_program)
        if sub_program:
            parser.add_argument("--fasta", help="Required. fasta file name.",required=True)
            parser.add_argument('--STAR_param', help=HELP_DICT['additional_param'], default="")

    @staticmethod
    def get_SA(fasta_size):
        """
        Get genomeSAindexNbases from fasta size. For small genomes, the parameter --genomeSAindexNbases must to be scaled down, with a typical value of min(14, log2(GenomeLength)/2 - 1). For example, for 1 megaBase genome, this is equal to 8, for 100 kiloBase genome, this is equal to 7.

        As mentioned in STAR manual, or 1 megaBase genome, this is equal to 9. But when the genome is relatively small, e.g. < 3kb,
        using large value may caused core dump error when mapping.

        Args:
            fasta_size: fasta size in bytes

        >>> Mkref.get_SA(10 ** 6)
        8
        >>> Mkref.get_SA(10 ** 5)
        7
        >>> Mkref.get_SA(10 ** 50)
        14
        """
        return min(int(math.log2(fasta_size) / 2 - 1), 14)

    def _get_SA(self):
        fasta_size = os.path.getsize(self.fasta)
        SA = self.get_SA(fasta_size)
        self.meta['genomeSAindexNbases'] = SA
        return SA

    @utils.add_log
    def build_star_index(self):
        SA = self._get_SA()
        cmd = (
            f'STAR \\\n'
            f'--runMode genomeGenerate \\\n'
            f'--runThreadN {self.thread} \\\n'
            f'--genomeDir ./ \\\n'
            f'--genomeFastaFiles {self.fasta} \\\n'
            f'--genomeSAindexNbases {SA} \\\n'
        )
        if self.STAR_param:
            cmd += (" " + self.args.STAR_param)
        sys.stderr.write(cmd + '\n')
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        self.build_star_index()


    
    


STEPS = [
    'mkref',
    'sample',
    'barcode',
    'cutadapt',
    'consensus',
    "star_virus",
    "count_virus",
    "filter_virus",
    "analysis_virus",
    "featureCounts",
    'count',
]
__ASSAY__ = 'capture_virus'

IMPORT_DICT = {
    'star_virus': 'celescope.rna_virus',
}



from celescope.tools.multi import Multi
from celescope.vdj.__init__ import __ASSAY__


class Multi_vdj(Multi):
    """

    ## Reference
    - Human
    ```
    mkdir -p /genome/vdj/human
    cd /genome/vdj/human
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/TR/TR{A,B}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/TR/TRBD.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/IG/IG{H,K,L}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/IG/IGHD.fasta
    celescope vdj mkref human TR
    celescope vdj mkref human IG
    ```

    - Mouse
    ```
    mkdir -p /genome/vdj/mouse
    cd /genome/vdj/mouse
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/TR/TR{A,B}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/TR/TRBD.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/IG/IG{H,K,L}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/IG/IGHD.fasta
    celescope vdj mkref mouse TR
    celescope vdj mkref mouse IG
    ```
 
    ## Usage

    Please note `multi_vdj` is only used for CDR3 vdj data(not full length). For full length vdj data, please use `multi_flv_CR` or  `multi_flv_trust4`.
    
    ```
    multi_vdj \\
        --mapfile ./vdj.mapfile \\
        --ref_path /genome/vdj/human/human_TR \\
        --species human \\
        --type TCR \\
        --thread 8 \\
        --mod shell
    ``` 

    """

    def consensus(self, sample):
        step = 'consensus'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
            f'--out_fasta '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)
        outfile = f'{self.outdir_dic[sample][step]}/{sample}_consensus.fasta'
        return outfile

    def mapping_vdj(self, sample):
        step = 'mapping_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        fasta = f'{self.outdir_dic[sample]["consensus"]}/{sample}_consensus.fasta'
        cmd = (
            f'{cmd_line} '
            f'--fasta {fasta} '
        )
        self.process_cmd(cmd, step, sample, m=15, x=self.args.thread)

    def count_vdj(self, sample):
        # count_vdj
        step = 'count_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        UMI_count_filter_file = (
            f'{self.outdir_dic[sample]["mapping_vdj"]}/{sample}_UMI_count_filtered.tsv'
        )
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--UMI_count_filter_file {UMI_count_filter_file} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=self.args.thread)


def main():
    multi = Multi_vdj(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()


import math
from collections import defaultdict
import operator,functools
import numpy as np

import plotly
import plotly.express as px
import plotly.graph_objects as go

from celescope.tools import utils

PLOTLY_CONFIG = {
    "displayModeBar": True,
    "staticPlot": False,
    "showAxisDragHandles": False,
    "modeBarButtons": [["toImage", "resetScale2d"]],
    "scrollZoom": False,
    "displaylogo": False
}

COLORS = px.colors.qualitative.Plotly + px.colors.qualitative.Alphabet

LAYOUT = {
    "height": 313,
    "width": 400,
    "margin": {
        "l": 45,
        "r": 35,
        "b": 30,
        "t": 30, }
}


def _round_float(x: float) -> float:
    """Round a float to the nearest value which can be represented with 4 decimal digits.

    In order to avoid representing floats with full precision, we convert them
    to a lower precision string and then convert that back to a float for use by `json.dumps`
    which will typically then output many fewer digits if possible.
    """
    return float(f"{x:.4g}")


def round_floats_in_list(x):
    """Lower the precision for a whole bunch of floats."""
    return [_round_float(x) for x in x]



class Plotly_plot():

    def __init__(self, df):
        self._df = df
        self._fig = None

    def plotly_plot(self):
        return plotly.offline.plot(
            self._fig,
            include_plotlyjs=False,
            output_type='div',
            config=PLOTLY_CONFIG
        )

    def get_plotly_div(self):

        return self.plotly_plot()


class Tsne_plot(Plotly_plot):

    def __init__(self, df_tsne, feature_name, discrete=True):
        super().__init__(df_tsne)

        self.feature_name = feature_name
        self.discrete = discrete
        title_feature_name = feature_name[0].upper() + feature_name[1:]
        self.title = f"t-SNE plot Colored by {title_feature_name}"

        self._layout = {}
        self._dot_size = 4
        self._df['size'] = self._dot_size
        self._df['barcode_index'] = list(range(1, len(self._df) + 1))
        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            'showgrid': True,
            'gridcolor': '#F5F5F5',
            'showline': False,
            'ticks': None,
            'zeroline': True,
            'zerolinecolor': 'black',
            'zerolinewidth': 0.7,
        }

        self.scatter_config = {
            'data_frame': df_tsne,
            'title': self.title,
            'x': self._str_coord1,
            'y': self._str_coord2,
            'size_max': self._dot_size,
            'hover_data': {
                self._str_coord1: False,
                self._str_coord2: False,
                self.feature_name: True,
                'barcode_index': True,
                'size': False,
            },
            'size': 'size',
            'opacity': 0.9,
            'color': self.feature_name,
            'color_discrete_sequence': COLORS,
            'color_continuous_scale': px.colors.sequential.Jet,
        }

    def set_color_scale(self, color_scale):
        self.scatter_config['color_continuous_scale'] = color_scale

    @utils.add_log
    def get_plotly_div(self):

        if self.discrete:
            self.discrete_tsne_plot()
        else:
            self.continuous_tsne_plot()
        self.update_fig()

        return self.plotly_plot()

    @utils.add_log
    def discrete_tsne_plot(self):

        sum_df = self._df.groupby([self.feature_name]).agg("count").iloc[:, 0]
        percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
        res_dict = defaultdict(int)
        res_list = []
        for cluster in sorted(self._df[self.feature_name].unique()):
            name = f"{cluster}({percent_df[cluster]}%)"
            res_dict[cluster] = name
            res_list.append(name)

        self._df[self.feature_name] = self._df[self.feature_name].map(res_dict)

        self._fig = px.scatter(
            **self.scatter_config,
            category_orders={self.feature_name: res_list}
        )

    @utils.add_log
    def continuous_tsne_plot(self):

        self._fig = px.scatter(
            **self.scatter_config,
        )

    def update_fig(self):
        self._fig.update_xaxes(
            title_text=self._str_coord1,
            **self.axes_config
        )

        self._fig.update_yaxes(
            title_text=self._str_coord2,
            **self.axes_config
        )

        self._fig.update_layout(
            self._layout,
            title={"text": self.title, "x": 0.5, "y": 0.95, "font": {"size": 15}},
            plot_bgcolor='#FFFFFF',
            hovermode="closest"
        )



class Tsne_dropdown_plot(Plotly_plot):

    def __init__(self, df_tsne,name, feature_name_list, discrete=False):
        super().__init__(df_tsne)

        self.name = name
        self.feature_name_list = feature_name_list
        self.discrete = discrete
        self.title = f"t-SNE plot Colored by {self.name}"
        self.x_pos = 0.1
        self._layout = {}

        self._buttons=[]
        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            'showgrid': True,
            'gridcolor': '#F5F5F5',
            'showline': False,
            'ticks': None,
            'zeroline': True,
            'zerolinecolor': 'black',
            'zerolinewidth': 0.7,
        }

        x_ = math.ceil(max(abs(self._df[self._str_coord1])))
        y_ = math.ceil(max(abs(self._df[self._str_coord2])))
        self.x_range = [-x_,x_]
        self.y_range = [-y_,y_]


    def get_plotly_div(self):

        if self.discrete:
            self.discrete_tsne_plot()
        else:
            self.continuous_tsne_plot()

        self.update_fig()

        return self.plotly_plot()

    def continuous_tsne_plot(self):
        self._fig = go.Figure()
        _num = len(self.feature_name_list)
        i=0

        max_num = math.ceil(max([self._df[_].max() for _ in self.feature_name_list]))
        split_num = math.ceil(max_num/5)
        tick_list = list(range(0,max_num,split_num))

        for feature_name in self.feature_name_list:
            df_tmp =  self._df.loc[:,[self._str_coord1,self._str_coord2,feature_name]]
            self._fig.add_trace(go.Scattergl(x=round_floats_in_list(df_tmp[self._str_coord1]),
                                           y=round_floats_in_list(df_tmp[self._str_coord2]),
                                           mode='markers',
                                           name = feature_name[0].upper() + feature_name[1:],
                                           showlegend=False,
                                           marker=go.scattergl.Marker(opacity=0.9,size=3,color=self._df[feature_name],
                                                                    cmax=max_num,
                                                                    cmin=0,
                                                                    colorscale=[[0,'LightGrey'],[1,'Red']],
                                                                    colorbar=go.scattergl.marker.ColorBar(
                                                                                                        tickmode='array',
                                                                                                        tickvals= tick_list,
                                                                                                        ticktext= tick_list,
                                                                                                        title="",
                                                                                                        titlefont={'size':11})),
                                           textposition='top center'
                                                ))
            visible_list=[False]*_num
            visible_list[i]=True
            i += 1
            button = dict(args=[{"visible":visible_list},],
              label=feature_name,
              method="restyle",
             )
            self._buttons.append(button)


    def discrete_tsne_plot(self):
        all_figures =[]
        num_all,last_num = 0,0
        for feature_name in self.feature_name_list:
            num_all += len(self._df[feature_name].unique())

        for feature_name in self.feature_name_list:
            df_sub = self._df.loc[:,['tSNE_1','tSNE_2',feature_name]]
            num_feature = len(df_sub[feature_name].unique())

            sum_df = df_sub.groupby(feature_name).agg("count").iloc[:, 0]
            percent_df = sum_df.transform(lambda x: round(x / sum(x) * 100, 2))
            res_dict = defaultdict(int)
            res_list = []
            for cluster in sorted(df_sub[feature_name].unique()):
                name = f"{cluster}({percent_df[cluster]}%)"
                res_dict[cluster] = name
                res_list.append(name)

            df_sub[feature_name] = df_sub[feature_name].map(res_dict)
            df_sub['size'] = 4

            scatter_config = {
                    'data_frame': df_sub,
                    'x': 'tSNE_1',
                    'y': 'tSNE_2',
                    'size_max':4,
                    'hover_data': {
                        'tSNE_1': False,
                        'tSNE_2': False,
                        feature_name: True,
                        'size': False,
                    },
                    'size': 'size',
                    'opacity': 0.9,
                    'color': feature_name,
                    'color_discrete_sequence': COLORS,
                    'color_continuous_scale': px.colors.sequential.Jet,
                }
            fig = px.scatter( **scatter_config,
                    category_orders={feature_name: res_list}
                )
            all_figures.append(fig)

            true_list = [True]*num_feature
            visible_list=[False]*num_all

            visible_list[last_num:num_feature+last_num]=true_list
            last_num += num_feature
            button = dict(args=[{"visible":visible_list},],
                          label=feature_name,
                          method="update",
            )
            self._buttons.append(button)

        self._fig = go.Figure(data=functools.reduce(operator.add,[_.data for _ in all_figures]))


    def update_fig(self):
        self._fig.update_xaxes(
            range=self.x_range,
            title_text=self._str_coord1,
            **self.axes_config
        )

        self._fig.update_yaxes(
            range=self.y_range,
            title_text=self._str_coord2,
            **self.axes_config
        )

        self._fig.update_layout(
            self._layout,
            updatemenus=[{'buttons':self._buttons,
                          'direction':'down',
                          'pad':{'r':8,'t':8},
                          'showactive':True,
                          'x':1.1,
                          'xanchor':'right',
                          'y':1.2,
                          'yanchor':'top'}],
            annotations=[{'text': 'TAG type:',
                          'showarrow': False,
                          'x': self.x_pos,
                          'xref': 'paper',
                          'xanchor': 'right',
                          'y': 1.15,
                          'yref': 'paper',
                          'font': {'size': 11},
                          'align': 'left'}],
            title={"text": self.title, "x": 0.5, "y": 0.95, "font": {"size": 15}},
            plot_bgcolor='#FFFFFF',
            hovermode="closest"
        )

# Add single tag plot
class Tsne_single_plot(Plotly_plot):

    def __init__(self, df_tsne, feature_name_list, analysis_dir):
        super().__init__(df_tsne)

        self.feature_name_list = feature_name_list
        self.analysis_dir = analysis_dir

        self._layout = {}

        self._str_coord1 = "tSNE_1"
        self._str_coord2 = "tSNE_2"
        self.axes_config = {
            'showgrid': True,
            'gridcolor': '#F5F5F5',
            'showline': False,
            'ticks': None,
            'zeroline': True,
            'zerolinecolor': 'black',
            'zerolinewidth': 0.7,
        }

        x_ = math.ceil(max(abs(self._df[self._str_coord1])))
        y_ = math.ceil(max(abs(self._df[self._str_coord2])))
        self.x_range = [-x_,x_]
        self.y_range = [-y_,y_]


    def get_plotly_div(self):

        max_num = math.ceil(max([self._df[_].max() for _ in self.feature_name_list]))
        split_num = math.ceil(max_num/5)
        tick_list = list(range(0,max_num,split_num))

        for feature_name in self.feature_name_list:
            self._fig = go.Figure()
            name = feature_name[0].upper() + feature_name[1:]
            title = f"t-SNE plot Colored by {name}"
            df_tmp =  self._df.loc[:,[self._str_coord1,self._str_coord2,feature_name]]
            df_tmp = df_tmp.loc[df_tmp[feature_name] != 0]
            self._fig.add_trace(go.Scatter(x=round_floats_in_list(df_tmp[self._str_coord1]),
                                            y=round_floats_in_list(df_tmp[self._str_coord2]),
                                            mode='markers',
                                            showlegend=False,
                                            marker=go.scatter.Marker(opacity=0.9,size=3,color=self._df[feature_name],
                                                                    cmax=max_num,
                                                                    cmin=0,
                                                                    colorscale=[[0,'LightGrey'],[1,'Red']],
                                                                    colorbar=go.scatter.marker.ColorBar(
                                                                                                        tickmode='array',
                                                                                                        tickvals= tick_list,
                                                                                                        ticktext= tick_list,
                                                                                                        title="",
                                                                                                        titlefont={'size':11})),
                                            textposition='top center'
                                                ))

            self.update_fig()
            self._fig.update_layout(title={"text": title, "x": 0.5, "y": 0.95, "font": {"size": 15}})
            self._fig.write_image(f"{self.analysis_dir}/{name}_tsne.png")

    def update_fig(self):
        self._fig.update_xaxes(
            range=self.x_range,
            title_text=self._str_coord1,
            **self.axes_config
        )

        self._fig.update_yaxes(
            range=self.y_range,
            title_text=self._str_coord2,
            **self.axes_config
        )

        self._fig.update_layout(
            self._layout,
            plot_bgcolor='#FFFFFF',
            hovermode="closest"
        )


class Bar_plot(Plotly_plot):

    def __init__(self, df_bar):
        super().__init__(df_bar)
        self.set_fig()

    def set_fig(self):
        self._fig = px.bar(
                    x=[str(i) for i in list(self._df.head(10).ClonotypeID)],
                    y=self._df.head(10).proportion.tolist(),
                    labels={'x': 'Clonotype ID', 'y': 'Fraction of Cells'},
                    width=700,
                    height=500
                    )
        self._fig.update_traces(marker_color='rgb(158,202,225)', marker_line_color='rgb(8,48,107)',
                   marker_line_width=1.5, opacity=0.6)
        self._fig.update_layout(title_text='Top 10 Clonotype Frequencies',
                   title={"x": 0.5, "y": 0.9, "font": {"size": 20, "family": "San Serif"}},
                   plot_bgcolor='#FFFFFF')


class Pie_plot(Plotly_plot):

    def __init__(self, df_region):
        super().__init__(df_region)
        self.set_fig()

    def set_fig(self):
        layout = {
            "height": 300,
            "width": 400,
            "margin": {
                "l": 50,
                "r": 35,
                "b": 10,
                "t": 10,
            }
        }
        self._fig = px.pie(
            self._df,
            names='regions',
            values='values',
        )
        self._fig.update_traces(textposition='none')
        self._fig.update_layout(layout)

class Line_plot(Plotly_plot):

    def __init__(self, df_line,title=None,x_title=None,y_title=None,y_range=None,section=True,):
        super().__init__(df_line)
        self.df_line = df_line
        self.title = title
        self.x_title = x_title
        self.y_title = y_title
        self.y_range = y_range
        self.section = section

        self.xaxes_config = {
            'showgrid': True,
            'gridcolor': '#F5F5F5',
            'linecolor':'black',
            'showline': True,
            'ticks': None,
            'tickmode':'linear',
            'tick0':0,
            'dtick':0.5,
        }

        if self.section:
            self.yaxes_config = {
                'showgrid': True,
                'gridcolor': '#F5F5F5',
                'linecolor':'black',
                'showline': True,
                'ticks': None,
                'rangemode':'tozero',
            }
        else:
            self.yaxes_config = {
                'showgrid': True,
                'gridcolor': '#F5F5F5',
                'linecolor':'black',
                'showline': True,
                'ticks': None,
                'range':self.y_range,
            }

        self.line_config = {
            'data_frame': df_line,
            'title': self.title,
            'x': self.x_title,
            'y': self.y_title,
        }

        self.line_plot()
        self.update_fig()

    @utils.add_log
    def line_plot(self):
        self._fig = px.line(
            **self.line_config,
        )

    def update_fig(self):
        self._fig.update_xaxes(
            **self.xaxes_config
        )

        self._fig.update_yaxes(
            **self.yaxes_config
        )

        self._fig.update_layout(
            LAYOUT,
            title={"x":0.5, "y":0.95, "font":{"size":15}},
            yaxis_zeroline = True,
            showlegend = False,
            plot_bgcolor = '#FFFFFF',
            hovermode = "closest"
        )


class Conversion_plot(Plotly_plot):

    def __init__(self, df_bar):
        super().__init__(df_bar)
        np.random.seed(1)
        self.values, bins = np.histogram(df_bar['cell_pct'], bins=200)
        self.bin_centers = (bins[:-1] + bins[1:]) / 2
        self.set_fig()
    
    def set_fig(self):
        self._fig = px.bar(x=self.bin_centers, y=self.values)
        self._fig.update_layout(yaxis_type='log')
        self._fig.update_xaxes(title='cell percent')
        self._fig.update_yaxes(title='count')
        self._fig.update_layout( plot_bgcolor='rgba(0,0,0,0)')


class Substitution_plot(Plotly_plot):

    def __init__(self, df_bar):
        super().__init__(df_bar)
        self.set_fig()

    def set_fig(self):

        num4sample = 0
        colors4sample = {}
        num4x = 0

        self._fig = go.Figure()

        for sample in self._df['sample'].unique():
            legend_show = True
            colors4sample[sample] = COLORS[num4sample]
            num4sample += 1
            flag_x = 'x' + str(num4x+1)
            df_plot = self._df[self._df['sample'] == sample]
            num4x += 1

            self._fig.add_trace(go.Bar(name=sample+'+',
                                 x=df_plot['sample'],
                                 y=df_plot['+'],
                                 legendgroup=sample,
                                 marker_color=colors4sample[sample],
                                 marker_line_color='#FFFFFF',
                                 showlegend=legend_show,
                                 xaxis=flag_x)
                          )
            self._fig.add_trace(go.Bar(name=sample+'-',
                                 x=df_plot['sample'],
                                 y=df_plot['-'],
                                 legendgroup=sample,
                                 showlegend=legend_show,
                                 marker_color=colors4sample[sample],
                                 marker_line_color='#FFFFFF',
                                 opacity=0.3,
                                 xaxis=flag_x)
                          )

        self._fig.update_layout(barmode='stack')

        per = 1/(num4x+1)
        gap4bar = per/len(self._df['sample'].unique())
        num4x = 0
        for _ in self._df['sample'].unique():
            if num4x == 0:
                flag_x = 'xaxis'
            else:
                flag_x = 'xaxis' + str(num4x+1)
            anchor_x = 'x'+str(num4x+1)
            num4x += 1
            self._fig['layout'][flag_x] = dict(domain=[per*num4x, per*(num4x+1)-gap4bar], anchor=anchor_x)

        self._fig.update_layout(plot_bgcolor='#FFFFFF')
        self._fig.update_xaxes(showgrid=False, linecolor='black', showline=True, ticks='outside', showticklabels=True)
        self._fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside')
        #width_num = 400 * (len(self._df['sample'].unique()) * len(self._df['sample'].unique())) / (5*12)  # set bar width
        self._fig.update_layout(height=500)#, width=width_num)
        self._fig.update_layout(legend=dict(orientation="h"))
        self._fig.update_layout(legend=dict(
            yanchor="top",y=1.3,
            xanchor="left",x=0.07,
            valign="top",
        ))
        self._fig.update_layout(
            yaxis_title="Rates of nucleotide substitution (%)"
        )


class Violin_plot(Plotly_plot):
    def __init__(self, df_vln, feature_name, color='#1f77b4', label=True):
        super().__init__(df_vln)
        self.feature_name = feature_name
        self.color = color
        self.label = label

        if self.label:
            self.y = "Fraction of labeled UMI"
        else:
            self.y = "Fraction of new RNA"
        self.set_fig()

    def set_fig(self):
        self._fig = go.Figure()
        self._fig.add_trace(
            go.Violin(y=self._df, box_visible=True, meanline_visible=True,
                    line_color='black', fillcolor=self.color, opacity=0.6,
                    x0=self.feature_name))
        self._fig.update_layout(yaxis_zeroline=True,  showlegend=False)
        self._fig.update_layout(plot_bgcolor='#FFFFFF')
        self._fig.update_xaxes(showgrid=False, linecolor='black', showline=True, ticks=None)
        self._fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside',
                            title_text=self.y, rangemode="tozero")


import glob
import gzip
import importlib
import logging
import os
import re
import subprocess
import time
import unittest
import json
import sys
from collections import Counter, defaultdict
from datetime import timedelta
from functools import wraps

from Bio.Seq import Seq
import pandas as pd
import pysam

from celescope.tools.__init__ import FILTERED_MATRIX_DIR_SUFFIX, BARCODE_FILE_NAME, OUTS_DIR
from celescope.__init__ import ROOT_PATH


def add_log(func):
    '''
    logging start and done.
    '''
    logFormatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    module = func.__module__
    name = func.__name__
    logger_name = f'{module}.{name}'
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

    consoleHandler = logging.StreamHandler(sys.stderr)
    consoleHandler.setFormatter(logFormatter)
    logger.addHandler(consoleHandler)

    @wraps(func)
    def wrapper(*args, **kwargs):

        logger.info('start...')
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        used = timedelta(seconds=end - start)
        logger.info('done. time used: %s', used)
        return result

    wrapper.logger = logger
    return wrapper

def generic_open(file_name, *args, **kwargs):
    if file_name.endswith('.gz'):
        file_obj = gzip.open(file_name, *args, **kwargs)
    else:
        file_obj = open(file_name, *args, **kwargs)
    return file_obj


class Gtf_dict(dict):
    '''
    key: gene_id
    value: gene_name
    If the key does not exist, return key. This is to avoid the error:
        The gtf file contains one exon lines with a gene_id, but do not contain a gene line with the same gene_id. FeatureCounts 
        work correctly under this condition, but the gene_id will not appear in the Gtf_dict.
    '''

    def __init__(self, gtf_file):
        super().__init__()
        self.gtf_file = gtf_file
        self.load_gtf()


    @add_log
    def load_gtf(self):
        """
        get gene_id:gene_name from gtf file
            - one gene_name with multiple gene_id: "_{count}" will be added to gene_name.
            - one gene_id with multiple gene_name: error.
            - duplicated (gene_name, gene_id): ignore duplicated records and print a warning.
            - no gene_name: gene_id will be used as gene_name.

        Returns:
            {gene_id: gene_name} dict
        """

        gene_id_pattern = re.compile(r'gene_id "(\S+)";')
        gene_name_pattern = re.compile(r'gene_name "(\S+)"')
        id_name = {}
        c = Counter()
        with generic_open(self.gtf_file, mode='rt') as f:
            for line in f:
                if not line.strip():
                    continue
                if line.startswith('#'):
                    continue
                tabs = line.split('\t')
                gtf_type, attributes = tabs[2], tabs[-1]
                if gtf_type == 'gene':
                    try:
                        gene_id = gene_id_pattern.findall(attributes)[-1]
                    except IndexError:
                        print(line)
                    gene_names = gene_name_pattern.findall(attributes)
                    if not gene_names:
                        gene_name = gene_id
                    else:
                        gene_name = gene_names[-1]
                    c[gene_name] += 1
                    if c[gene_name] > 1:
                        if gene_id in id_name:
                            assert id_name[gene_id] == gene_name, (
                                'one gene_id with multiple gene_name '
                                f'gene_id: {gene_id}, '
                                f'gene_name this line: {gene_name}'
                                f'gene_name previous line: {id_name[gene_id]}'
                            )
                            self.load_gtf.logger.warning(
                                'duplicated (gene_id, gene_name)'
                                f'gene_id: {gene_id}, '
                                f'gene_name {gene_name}'
                            )
                            c[gene_name] -= 1
                        else:
                            gene_name = f'{gene_name}_{c[gene_name]}'
                    id_name[gene_id] = gene_name
        self.update(id_name)

    def __getitem__(self, key):
        '''if key not exist, return key'''
        return dict.get(self, key, key)


def read_one_col(file):
    """
    Read file with one column. Strip each line.
    Returns col_list, line number
    """
    df = pd.read_csv(file, header=None)
    col1 = list(df.iloc[:, 0])
    col1 = [item.strip() for item in col1]
    num = len(col1)
    return col1, num


def get_bed_file_path(panel):
    bed_file_path = f'{ROOT_PATH}/data/snp/panel/{panel}.bed'
    if not os.path.exists(bed_file_path):
        return None
    else:
        return bed_file_path


def get_gene_region_from_bed(panel):
    """
    Returns 
    - genes
    - position_df with 'Chromosome', 'Start', 'End'
    """
    file_path = get_bed_file_path(panel)
    bed_file_df = pd.read_table(file_path,
                                usecols=[0, 1, 2, 3],
                                names=['Chromosome', 'Start', 'End', 'Gene'],
                                sep="\t")
    position_df = bed_file_df.loc[:, ['Chromosome', 'Start', 'End']]
    genes = set(bed_file_df.loc[:, 'Gene'].to_list())
    return genes, position_df


def read_fasta(fasta_file, equal=False):
    """
    Args:
        equal: if True, seq in fasta must have equal length
    Returns:
        {seq_id: seq} dict
    """
    fa_dict = {}
    length = None
    with pysam.FastxFile(fasta_file) as infile:
        for index, record in enumerate(infile):
            seq = record.sequence
            if index == 0:
                length = len(seq)
            if equal:
                if length != len(seq):
                    raise Exception(f"{fasta_file} have different seq length")
            fa_dict[record.name] = seq
    return fa_dict, length


def hamming_correct(string1, string2):
    threshold = len(string1) / 10 + 1
    if hamming_distance(string1, string2) < threshold:
        return True
    return False


def hamming_distance(string1, string2):
    distance = 0
    length = len(string1)
    length2 = len(string2)
    if (length != length2):
        raise Exception(f"string1({length}) and string2({length2}) do not have same length")
    for i in range(length):
        if string1[i] != string2[i]:
            distance += 1
    return distance


def format_number(number: int) -> str:
    return format(number, ",")


def genDict(dim=3, valType=int):
    if dim == 1:
        return defaultdict(valType)
    else:
        return defaultdict(lambda: genDict(dim - 1, valType=valType))


class MultipleFileFoundError(Exception):
    pass

def glob_file(pattern_list: list):
    """
    glob file among pattern list
    Returns:
        PosixPath object
    Raises:
        FileNotFoundError: if no file found
        MultipleFileFound: if more than one file is found
    """
    if not isinstance(pattern_list, list):
        raise TypeError('pattern_list must be a list')

    match_list = []
    for pattern in pattern_list:
        files = glob.glob(pattern)
        if files:
            for f in files:
                match_list.append(f)
    
    if len(match_list) == 0:
        raise FileNotFoundError(f'No file found for {pattern_list}')
    
    if len(match_list) > 1:
        raise MultipleFileFoundError(
            f'More than one file found for pattern: {pattern_list}\n'
            f'File found: {match_list}'
        )
    
    return match_list[0]


def get_matrix_file_path(matrix_dir, file_name):
    """
    compatible with gzip file
    """
    file_path_list = [f'{matrix_dir}/{file_name}', f'{matrix_dir}/{file_name}.gz']
    for file_path in file_path_list:
        if os.path.exists(file_path):
            return file_path


@add_log
def get_barcode_from_matrix_dir(matrix_dir):
    """
    Returns:
        match_barcode: list
        no_match_barcode: int
    """
  
    match_barcode_file = get_matrix_file_path(matrix_dir, BARCODE_FILE_NAME)
    match_barcode, n_match_barcode = read_one_col(match_barcode_file)

    return match_barcode, n_match_barcode


@add_log
def get_matrix_dir_from_match_dir(match_dir):
    """
    Returns:
        matrix_dir: PosixPath object
    """
    matrix_dir_pattern_list = []
    for matrix_dir_suffix in FILTERED_MATRIX_DIR_SUFFIX:
        matrix_dir_pattern_list.append(f"{match_dir}/{OUTS_DIR}/*{matrix_dir_suffix}")
  
    matrix_dir = glob_file(matrix_dir_pattern_list)
    get_matrix_dir_from_match_dir.logger.info(f"Matrix_dir :{matrix_dir}")

    return matrix_dir


@add_log
def get_barcode_from_match_dir(match_dir):
    '''
    multi version compatible
    Returns:
        match_barcode: list
        no_match_barcode: int
    '''
    matrix_dir = get_matrix_dir_from_match_dir(match_dir)
    return get_barcode_from_matrix_dir(matrix_dir)


@add_log
def parse_match_dir(match_dir):
    '''
    return dict
    keys: 'match_barcode', 'n_match_barcode', 'matrix_dir', 'tsne_coord'
    '''
    match_dict = {}

    pattern_dict = {
        'tsne_coord': [f'{match_dir}/{OUTS_DIR}/tsne_coord.tsv'],
        'markers': [f'{match_dir}/{OUTS_DIR}/markers.tsv'],
        'h5ad': [f'{match_dir}/{OUTS_DIR}/rna.h5ad'],
    }

    for file_key in pattern_dict:
        file_pattern= pattern_dict[file_key]
        try:
            match_file = glob_file(file_pattern)
        except FileNotFoundError:
            parse_match_dir.logger.warning(f"No {file_key} found in {match_dir}")
        else:
            match_dict[file_key] = match_file

    match_dict['matrix_dir'] = get_matrix_dir_from_match_dir(match_dir)
    match_barcode, n_match_barcode = get_barcode_from_match_dir(match_dir)
    match_dict['match_barcode'] = match_barcode
    match_dict['n_match_barcode'] = n_match_barcode

    return match_dict


def fastq_line(name, seq, qual):
    return f'@{name}\n{seq}\n+\n{qual}\n'


def fasta_line(name, seq):
    return f'>{name}\n{seq}\n'


def find_assay_init(assay):
    init_module = importlib.import_module(f"celescope.{assay}.__init__")
    return init_module


def find_step_module(assay, step):
    file_path_dict = {
        'assay': f'{ROOT_PATH}/{assay}/{step}.py',
        'tools': f'{ROOT_PATH}/tools/{step}.py',
    }

    init_module = find_assay_init(assay)
    if os.path.exists(file_path_dict['assay']):
        step_module = importlib.import_module(f"celescope.{assay}.{step}")
    elif hasattr(init_module, 'IMPORT_DICT') and step in init_module.IMPORT_DICT:
        module_path = init_module.IMPORT_DICT[step]
        step_module = importlib.import_module(f"{module_path}.{step}")
    elif os.path.exists(file_path_dict['tools']):
        step_module = importlib.import_module(f"celescope.tools.{step}")
    else:
        raise ModuleNotFoundError(f"No module found for {assay}.{step}")

    return step_module


def find_step_module_with_folder(assay, step):
    init_module = find_assay_init(assay)
    folder = ""
    try:
        step_module = importlib.import_module(f"celescope.{assay}.{step}")
        folder = assay
    except ModuleNotFoundError:
        try:
            step_module = importlib.import_module(f"celescope.tools.{step}")
            folder = 'tools'
        except ModuleNotFoundError:
            module_path = init_module.IMPORT_DICT[step]
            step_module = importlib.import_module(f"{module_path}.{step}")
            folder = module_path.split('.')[1]

    return step_module, folder


def sort_bam(input_bam, output_bam, threads=1, by='pos'):
    cmd = (
        f'samtools sort {input_bam} '
        f'-o {output_bam} '
        f'--threads {threads} '
        '2>&1 '
    )
    if by == "name":
        cmd += " -n"
    subprocess.check_call(cmd, shell=True)


def index_bam(input_bam):
    cmd = f"samtools index {input_bam} 2>&1 "
    subprocess.check_call(cmd, shell=True)


def add_tag(seg, id_name, correct_dict):
    """
    Args:
        seg: pysam bam segment
        id_name: {gene_id: gene_name}
        correct_dict: {low_seq: high_seq}

    Returns:
        seg with tag added

    """
    attr = seg.query_name.split(':')
    barcode = attr[0]
    umi = attr[1]
    seg.set_tag(tag='CB', value=barcode, value_type='Z')
    if umi in correct_dict:
        umi = correct_dict[umi]
    seg.set_tag(tag='UB', value=umi, value_type='Z')
    # assign to some gene
    if seg.has_tag('XT'):
        gene_id = seg.get_tag('XT')
        # if multi-mapping reads are included in original bam,
        # there are multiple gene_ids
        if ',' in gene_id:
            gene_name = [id_name[i] for i in gene_id.split(',')]
            gene_name = ','.join(gene_name)
        else:
            gene_name = id_name[gene_id]
        seg.set_tag(tag='GN', value=gene_name, value_type='Z')
        seg.set_tag(tag='GX', value=gene_id, value_type='Z')

    return seg


def check_mkdir(dir_name):
    """if dir_name is not exist, make one"""
    if not os.path.exists(dir_name):
        os.system(f"mkdir -p {dir_name}")


class Samtools():
    def __init__(self, in_bam, out_bam, threads=1, debug=False):
        self.in_bam = in_bam
        self.out_bam = out_bam
        self.threads = threads
        self.temp_sam_file = f"{self.out_bam}_sam.temp"
        self.debug = debug

    @add_log
    def samtools_sort(self, in_file, out_file, by='pos'):
        cmd = f"samtools sort {in_file} -o {out_file} --threads {self.threads}"
        if by == "name":
            cmd += " -n"
        self.samtools_sort.logger.debug(cmd)
        subprocess.check_call(cmd, shell=True)

    @add_log
    def samtools_index(self, in_file):
        cmd = f"samtools index {in_file}"
        self.samtools_index.logger.debug(cmd)
        subprocess.check_call(cmd, shell=True)

    def sort_bam(self, by='coord'):
        """sort in_bam"""
        self.samtools_sort(self.in_bam, self.out_bam, by=by)

    def index_bam(self):
        """index out_bam"""
        self.samtools_index(self.out_bam)

    @add_log
    def add_tag(self, gtf_file):
        """
        - CB cell barcode
        - UB UMI
        - GN gene name
        - GX gene id
        """
        gtf_dict = Gtf_dict(gtf_file)

        with pysam.AlignmentFile(self.in_bam, "rb") as original_bam:
            header = original_bam.header
            with pysam.AlignmentFile(self.temp_sam_file, "w", header=header) as temp_sam:
                for read in original_bam:
                    attr = read.query_name.split(':')
                    barcode = attr[0]
                    umi = attr[1]
                    read.set_tag(tag='CB', value=barcode, value_type='Z')
                    read.set_tag(tag='UB', value=umi, value_type='Z')
                    # assign to some gene
                    if read.has_tag('XT'):
                        gene_id = read.get_tag('XT')
                        # if multi-mapping reads are included in original bam,
                        # there are multiple gene_ids
                        if ',' in gene_id:
                            gene_name = [gtf_dict[i] for i in gene_id.split(',')]
                            gene_name = ','.join(gene_name)
                        else:
                            gene_name = gtf_dict[gene_id]
                        read.set_tag(tag='GN', value=gene_name, value_type='Z')
                        read.set_tag(tag='GX', value=gene_id, value_type='Z')
                    temp_sam.write(read)

    @add_log
    def add_RG(self, barcodes):
        """
        barcodes list
        """

        with pysam.AlignmentFile(self.in_bam, "rb") as original_bam:
            header = original_bam.header.to_dict()
            header['RG'] = []
            for index, barcode in enumerate(barcodes):
                header['RG'].append({
                    'ID': barcode,
                    'SM': index + 1,
                })

            with pysam.AlignmentFile(self.temp_sam_file, "w", header=header) as temp_sam:
                for read in original_bam:
                    read.set_tag(tag='RG', value=read.get_tag('CB'), value_type='Z')
                    temp_sam.write(read)

    def temp_sam2bam(self, by=None):
        self.samtools_sort(self.temp_sam_file, self.out_bam, by=by)
        self.rm_temp_sam()

    def rm_temp_sam(self):
        cmd = f"rm {self.temp_sam_file}"
        subprocess.check_call(cmd, shell=True)


def read_CID(CID_file):
    """
    return df_index, df_valid
    """
    df_index = pd.read_csv(CID_file, sep='\t', index_col=0).reset_index()
    df_valid = df_index[df_index['valid'] == True]
    return df_index, df_valid


def get_assay_text(assay):
    """
    Deprecated
    add sinlge cell prefix
    deprecated
    """
    return 'Single-cell ' + assay


def check_arg_not_none(args, arg_name):
    """
    check if args.arg_name is not None
    Args:
        args: argparser args
        arg_name: argparser arg name
    Return:
        bool
    """
    arg_value = getattr(args, arg_name, None)
    if arg_value and arg_value.strip() != 'None':
        return True
    else:
        return False


def reverse_complement(seq):
    """Reverse complementary sequence

    :param original seq
    :return Reverse complementary sequence
    """
    return str(Seq(seq).reverse_complement())

def get_fastx_read_number(fastx_file):
    """
    get read number using pysam
    """
    n = 0
    with pysam.FastxFile(fastx_file) as f:
        for _ in f:
            n += 1
    return n

@add_log
def dump_dict_to_json(d, json_file):
    with open(json_file, 'w') as f:
        json.dump(d, f, indent=4)

@add_log
def barcode_list_stamp(barcode_list, cut=500):
    bc_list, bc_num = read_one_col(barcode_list)
    n, m = 0, 0
    stamp = defaultdict(list)
    for i in bc_list:
        m += 1
        if m <= cut:
            stamp[n].append(i)
        else:
            n += 1
            m = 1
            stamp[n].append(i)
    return stamp, bc_num


class Test_utils(unittest.TestCase):
    def test_gtf_dict(self):
        import tempfile
        fp = tempfile.NamedTemporaryFile(suffix='.gtf')
        fp.write(b'1\tprocessed_transcript\tgene\t11869\t14409\t.\t+\t.\tgene_id "gene_id_with_gene_name"; gene_name "gene_name";\n')
        fp.write(b'1\tprocessed_transcript\tgene\t11869\t14409\t.\t+\t.\tgene_id "gene_id_without_gene_name"; \n')
        fp.seek(0)
        gtf_dict = Gtf_dict(fp.name)
        self.assertEqual(gtf_dict['gene_id_with_gene_name'], 'gene_name')
        self.assertEqual(gtf_dict['gene_id_without_gene_name'], 'gene_id_without_gene_name')
        self.assertEqual(gtf_dict['gene_id_not_exist'], 'gene_id_not_exist')
        fp.close()

if __name__ == '__main__':
    unittest.main()


import argparse
import inspect
import os
from collections import OrderedDict

from celescope.tools import utils
from celescope.celescope import ArgFormatter
from celescope.__init__ import ASSAY_LIST, RELEASED_ASSAYS

PRE_PROCESSING_STEPS = ('sample', 'barcode', 'cutadapt')
DOCS_DIR = f'docs/'
MANUAL = f'{DOCS_DIR}/manual.md'

def parse_docstring_to_dict(docstring):
    """
    Parse docstring(markown) to dict
    Return:
        dict {title: content}
    """

    doc_dict = {}
    title = None
    for line in docstring.split('\n'):
        if line.find('##') != -1:
            title = line.replace('##', '').strip().lower()
            doc_dict[title] = ""
        else:
            if title:
                doc_dict[title] += line + '\n'

    return doc_dict



def get_argument_docs_from_parser(parser):
    argument_docs = ""
    for argument in parser._option_string_actions:
        if not argument in ['-h', '--help']:
            help_msg = parser._option_string_actions[argument].help
            if help_msg:
                help_msg = help_msg.strip()
                # if end with code segment, do not add dot
                if help_msg[-1] != '.' and help_msg[-3:] != "```":
                    help_msg += '.'
            argument_docs += (f'`{argument}` {help_msg}\n\n')
    argument_docs = "## Arguments\n" + argument_docs
    return argument_docs


def get_class_docs(step_module):
    """
    Return:
        class_docs: str
        doc_dict: dict
    """
    # case ignored
    titles = ("features", "output", "usage")

    for child in inspect.getmembers(step_module, inspect.isclass):
        """Filter out class not defined in step_module"""
        class_obj = child[1]
        if class_obj.__module__ != step_module.__name__:
            continue
        doc = inspect.getdoc(class_obj)
        
        # only write class docs when the doc contains at least one title
        write_bool = False
        if doc:
            doc_dict = parse_docstring_to_dict(doc)
            for title in titles:
                if title in doc_dict:
                    write_bool = True
                    break

        if write_bool:
            class_docs = doc + '\n'
            return class_docs, doc_dict

    return "", {}


class Docs():
    def __init__(self, assay):
        self.assay = assay

        init_module = utils.find_assay_init(assay)
        self.steps = init_module.STEPS.copy()
        self.multi_step = f'multi_{assay}'

        self.out_md_dict = {}
        self.relative_md_path = {}
        self.release_bool = self.assay in RELEASED_ASSAYS

        # order: step
        self.output_file_doc_dict = OrderedDict()
        self.feature_doc_dict = OrderedDict()

        assay_dir = f'docs/{assay}'
        if not os.path.exists(assay_dir):
            os.system(f'mkdir -p {assay_dir}') 
        tools_dir = 'docs/tools'
        if not os.path.exists(tools_dir):
            os.system(f'mkdir {tools_dir}')

        # manual
        self.manual_lines = []


    @utils.add_log
    def get_argument_docs(self, step, step_module):
        self.get_argument_docs.logger.info(step)
        if step.startswith("multi"):
            multi_class = getattr(step_module, f'Multi_{self.assay}')
            multi_obj = multi_class(self.assay)
            argument_docs = get_argument_docs_from_parser(multi_obj.parser)
        else:
            parser = argparse.ArgumentParser(description='CeleScope', formatter_class=ArgFormatter)
            func_opts = getattr(step_module, f"get_opts_{step}")
            func_opts(parser, sub_program=True)
            argument_docs = get_argument_docs_from_parser(parser)
        return argument_docs

    def get_multi_feature_doc(self):
        multi_feature_doc = "## Features\n"
        for step in self.feature_doc_dict:
            multi_feature_doc += f"### {step}\n"
            multi_feature_doc += self.feature_doc_dict[step]
            multi_feature_doc += '\n'

        return multi_feature_doc

    def get_multi_output_doc(self):
        multi_output_doc = "## Output files\n"
        for step in self.output_file_doc_dict:
            multi_output_doc += f"### {step}\n"
            multi_output_doc += self.output_file_doc_dict[step]
            multi_output_doc += '\n'

        return multi_output_doc

    def write_step_doc(self, step):
        """
        folder: docs/folder/*.md
        """
        step_module = utils.find_step_module(self.assay, step)
        folder = step_module.__name__.split('.')[1]
        self.out_md_dict[step] = f'docs/{folder}/{step}.md'
        self.relative_md_path[step] = f'{folder}/{step}.md'

        class_docs, doc_dict = get_class_docs(step_module)
        if 'output' in doc_dict:
            self.output_file_doc_dict[step] = doc_dict['output']
        if 'features' in doc_dict:
            self.feature_doc_dict[step] = doc_dict['features']
        argument_docs = self.get_argument_docs(step, step_module)

        with open(self.out_md_dict[step], 'w') as out_file:
            out_file.write(class_docs)
            if step.startswith("multi"):
                out_file.write(self.get_multi_feature_doc())
                out_file.write(self.get_multi_output_doc())
            out_file.write(argument_docs)

    def run(self):
        if self.release_bool:
            self.manual_lines.append(f'## {self.assay}\n')       

        for step in self.steps:
            self.write_step_doc(step)
            if self.release_bool:
                self.add_step_in_manual(step)

        # write multi last because multi need to collect output files from each step
        if self.assay != 'utils':
            self.write_step_doc(self.multi_step)
            if self.release_bool:
                self.add_step_in_manual(self.multi_step)

        self.write_manual()


    def add_step_in_manual(self, step):
        """
        - [mkref](rna/mkref.md)
        """
        if not step in PRE_PROCESSING_STEPS:
            self.manual_lines.append(f'- [{step}]({self.relative_md_path[step]})\n')

    def write_manual(self):
        with open(MANUAL, 'a') as writer:
            for line in self.manual_lines:
                writer.write(line)


def remove_old_assays():
    contents = []
    with open(MANUAL, 'r') as reader:
        for line in reader:
            contents.append(line)
            if line.find("tools/cutadapt.md") != -1:
                break
    
    with open(MANUAL, 'w') as writer:
        for line in contents:
            writer.write(line)


def main():

    remove_old_assays()
    
    for assay in ASSAY_LIST:
        docs_obj = Docs(assay)
        docs_obj.run()


if __name__ == "__main__":
    main()



import argparse

# On some system, import scanpy will report "ImportError: dlopen: cannot load any more object with static TLS" when scanpy import sklearn
# as described in 
# https://github.com/singleron-RD/CeleScope/issues/127
# https://github.com/scikit-learn/scikit-learn/issues/14485
# https://github.com/scverse/scanpy/issues/1121
# https://github.com/pytorch/pytorch/issues/2575
# After testing, the version of glibc and sklearn is not the reason
# Import scanpy before scipy.stats(used in tools/count.py) will do the trick.
# However, this approach is only tested on one system.
import scanpy # noqa # pylint: disable=unused-import

from celescope.tools import utils
from celescope.__init__ import __VERSION__, ASSAY_LIST


class ArgFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter):
    pass


def main():
    """celescope cli
    """
    parser = argparse.ArgumentParser(description='CeleScope', formatter_class=ArgFormatter)
    parser.add_argument('-v', '--version', action='version', version=__VERSION__)
    subparsers = parser.add_subparsers(dest='subparser_assay')

    for assay in ASSAY_LIST:
        subparser_1st = subparsers.add_parser(assay)

        # add 2ed subparser
        subparser_2nd = subparser_1st.add_subparsers()

        # import STEPS
        init_module = utils.find_assay_init(assay)
        STEPS = init_module.STEPS

        for step in STEPS:
            # import function and opts
            step_module = utils.find_step_module(assay, step)
            func = getattr(step_module, step)
            func_opts = getattr(step_module, f"get_opts_{step}")
            parser_step = subparser_2nd.add_parser(step, formatter_class=ArgFormatter)
            func_opts(parser_step, sub_program=True)
            parser_step.set_defaults(func=func)

    args = parser.parse_args()
    if len(args.__dict__) <= 1:
        # No arguments or subcommands were given.
        parser.print_help()
        parser.exit()
    else:
        args.func(args)


if __name__ == '__main__':
    main()




from celescope.citeseq.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_citeseq(Multi):
    """
    ## Usage
    ```
    multi_citeseq \\
        --mapfile ./test.mapfile \\
        --barcode_fasta ./CLindex_TAG.fasta \\
        --fq_pattern L25C15 \\
        --mod shell
    ```
    """

    def mapping_tag(self, sample):
        step = 'mapping_tag'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def count_cite(self, sample):

        step = 'count_cite'
        cmd_line = self.get_cmd_line(step, sample)
        read_count_file = f'{self.outdir_dic[sample]["mapping_tag"]}/{sample}_read_count.tsv'
        cmd = (
            f'{cmd_line} '
            f'--read_count_file {read_count_file} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def analysis_cite(self, sample):

        step = 'analysis_cite'
        cmd_line = self.get_cmd_line(step, sample)
        tsne_coord = f'{self.outdir_dic[sample]["count_cite"]}/{sample}_tsne_coord.tsv'
        cmd = (
            f'{cmd_line} '
            f'--tsne_coord {tsne_coord} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)


def main():
    multi = Multi_citeseq(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



import pandas as pd
from celescope.tools import utils
from celescope.tools import count as super_count
from celescope.tools.matrix import ROW

class Count(super_count.Count):
    """
    ## Features
    - Generate expression matrix.
    - Well statstic.

    ## Output
    - `{sample}_raw_feature_bc_matrix` The expression matrix of all detected barcodes in [Matrix Market Exchange Formats](
        https://math.nist.gov/MatrixMarket/formats.html). 
    - `{sample}_count_detail.txt.gz` 4 columns: 
        - barcode  
        - gene ID  
        - UMI count  
        - read_count  
    - `{sample}_counts.txt` 6 columns:
        - Barcode: barcode sequence
        - read: read count of each barcode
        - UMI: UMI count for each barcode
        - geneID: gene count for each barcode
        - mark: cell barcode or backgound barcode.
            `CB` cell  
            `UB` background  
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
        self.raw_count_file = f'{self.outdir}/{self.sample}_counts.txt'
        self.marked_count_file = f'{self.outdir}/{self.sample}_counts_report.txt'
        self.umi_cutoff = args.umi_cutoff
        self.read_cutoff = args.read_cutoff
        self.gene_cutoff = args.gene_cutoff
        self._table_id = self.assay

    
    @utils.add_log
    def run(self):
        ## output exprssion matrix
        df = pd.read_table(self.args.count_detail, header=0, dtype={'geneID':str}, index_col=[0,1])
        self.write_sparse_matrix(df, self.raw_matrix_dir)
        ## output stats
        df_bc = self.get_df_bc(df)
        df_bc.index.name = 'Well'
        sort_col = ['UMI', 'read', ROW]
        df_bc = df_bc[sort_col]
        df_bc.columns = ['UMI', 'read', 'gene']
        df_bc.to_csv(self.raw_count_file, sep='\t')
        df_bc_valid = df_bc[ (df_bc['UMI']>=self.umi_cutoff) & (df_bc['read']>=self.read_cutoff) & (df_bc['gene']>=self.gene_cutoff) ]
        df_bc_valid.to_csv(self.marked_count_file, sep='\t')
        ## report
        self.add_count_metrics(df_bc_valid)

    @utils.add_log
    def add_count_metrics(self, df):
        out_condi = ' '
        out_condi += f'umi>={self.umi_cutoff} ' if self.umi_cutoff>0 else ''
        out_condi += f'read>={self.read_cutoff} ' if self.read_cutoff>0 else ''
        out_condi += f'gene>={self.gene_cutoff} ' if self.gene_cutoff>0 else ''
        self.add_help_content(
            name='Well:', content=f'The 1st column in the table is the well barcode sequence. Only output wells that meet the conditions({out_condi}).')
        stats = df.describe()
        stats.columns = ['UMI', 'Reads', 'Genes']
        for item in ['Reads', 'UMI', 'Genes']:
            self.add_metric(
                name=f'Median {item} per Well',
                value=int(stats.loc['50%',item]),
                help_info=''
            )
        for item in ['Reads', 'UMI', 'Genes']:
            self.add_metric(
                name=f'Mean {item} per Well',
                value=int(stats.loc['mean',item]),
                help_info=''
            )
        table_dict = self.get_table_dict(
            title='Detailed information per Well',
            table_id=self._table_id,
            df_table=df.reset_index(),
        )
        self.add_data(table_dict=table_dict)



def count(args):
    with Count(args, display_title="Wells") as runner:
        runner.run()

def get_opts_count(parser, sub_program):
    parser.add_argument('--umi_cutoff', default=500, type=int,
        help='If the UMI number exceeds the threshold, it is considered a valid well and reported.'
    )
    parser.add_argument('--gene_cutoff', default=0, type=int,
        help='If the gene number exceeds the threshold, it is considered a valid well and reported.'
    )
    parser.add_argument('--read_cutoff', default=0, type=int,
        help='If the read number exceeds the threshold, it is considered a valid well and reported.'
    )
    super_count.get_opts_count(parser, sub_program)




from celescope.tools.capture.count_bam import Count_bam, get_opts_count_bam


def get_opts_count_virus(parser, sub_program):
    get_opts_count_bam(parser, sub_program)


class Count_virus(Count_bam):
    """
    ## Features

    - Taking the bam file as input, count the number of UMIs and reads mapped to the viral genome.

    ## Output
    - {sample_raw_read_count.json} : barcode - UMI - raw_reads_count
    """



def count_virus(args):

    with Count_virus(args) as runner:
        runner.run()



import scanpy as sc
import numpy as np
import pandas as pd

from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.rna.mkref import Mkref_rna
from celescope.tools.step import Step, s_common

# markers adjust p_value
PVAL_CUTOFF = 0.05
# scanpy mitochonrial variable name
MITO_VAR = 'mito'
NORMALIZED_LAYER = 'normalised'
RESOLUTION = 1.2
N_PCS = 25
MITO_GENE_PERCENT_LIST = [5, 10, 15, 20, 50]
# output marker top n in html
MARKER_TOP_N = 100


def read_tsne(tsne_file):
    df = pd.read_csv(tsne_file, sep='\t', index_col=0)
    df.index.names = ['barcode']
    return df

def format_df_marker(df_marker):

    avg_logfc_col = "avg_log2FC"  # seurat 4
    if "avg_logFC" in df_marker.columns:  # seurat 2.3.4
        avg_logfc_col = "avg_logFC"
    df_marker = df_marker.loc[:,
                                ["cluster", "gene", avg_logfc_col, "pct.1", "pct.2", "p_val_adj"]
                                ]
    df_marker["cluster"] = df_marker["cluster"].apply(lambda x: f"cluster {x}")
    df_marker = df_marker[df_marker["p_val_adj"] < PVAL_CUTOFF]

    return df_marker

def get_opts_analysis(parser, sub_program):
    
    parser.add_argument('--genomeDir', help=HELP_DICT['genomeDir'], required=True)
    if sub_program:
        parser.add_argument(
            '--matrix_file',
            help='Required. Matrix directory path.',
            required=True,
        )
        parser = s_common(parser)


class Scanpy_wrapper(Step):

    def __init__(self, args, display_title=None):

        super().__init__(args, display_title=display_title)

        # data
        self.adata = sc.read_10x_mtx(
            args.matrix_file,  
            var_names='gene_symbols',
        )
        self.adata.layers['raw'] = self.adata.X.copy()
        self.mt_gene_list = Mkref_rna.get_config(args.genomeDir)['files']['mt_gene_list']

        # out
        self.df_marker_file = f'{self.outdir}/markers.tsv'
        self.df_marker_raw_file = f'{self.outdir}/markers_raw.tsv'
        self.df_tsne_file = f'{self.outdir}/tsne_coord.tsv'
        self.h5ad_file = f'{self.outdir}/rna.h5ad'
        self.outs = [self.df_marker_file, self.df_tsne_file, self.h5ad_file]

   
    @utils.add_log
    def calculate_qc_metrics(self):

        if self.mt_gene_list:
            mito_genes, _ = utils.read_one_col(self.mt_gene_list)
            self.adata.var[MITO_VAR] = self.adata.var_names.map(lambda x:True if x in mito_genes else False)
            # if not astype(bool), it will be type object and raise an error
            # https://github.com/theislab/anndata/issues/504
            self.adata.var[MITO_VAR] = self.adata.var[MITO_VAR].astype(bool)
        else:
            self.adata.var[MITO_VAR] = self.adata.var_names.str.upper().str.startswith('MT-')

        sc.pp.calculate_qc_metrics(
            self.adata,
            qc_vars=[MITO_VAR], 
            percent_top=None,
            use_raw=False,
            log1p=False, 
            inplace=True
        )

    @utils.add_log
    def write_mito_stats(self):

        mt_pct_var = f'pct_counts_{MITO_VAR}'
        total_cell_number = self.adata.n_obs

        for mito_gene_percent in MITO_GENE_PERCENT_LIST:
            cell_number = sum(self.adata.obs[mt_pct_var] > mito_gene_percent)
            fraction = round(cell_number / total_cell_number * 100, 2)
            self.add_metric(
                name=f'Fraction of cells have mito gene percent>{mito_gene_percent}%',
                value=f'{fraction}%',
            )


    @utils.add_log
    def normalize(self):
        """
        sc.pp.normalize_per_cell() and sc.pp.log1p()
        """

        sc.pp.normalize_total(
            self.adata,
            target_sum=1e4,
            inplace=True,
        )
        sc.pp.log1p(
            self.adata,
        )
        self.adata.layers[NORMALIZED_LAYER] = self.adata.X

    @utils.add_log
    def hvg(self):
        """
        Wrapper function for sc.highly_variable_genes()
        """
        sc.pp.highly_variable_genes(
            self.adata,
            layer=None,
            n_top_genes=None,
            min_disp=0.5,
            max_disp=np.inf,
            min_mean=0.0125,
            max_mean=3,
            span=0.3,
            n_bins=20,
            flavor='seurat',
            subset=False,
            inplace=True,
            batch_key=None,
            check_values=True
        )        

    @utils.add_log
    def scale(self):
        """
        Wrapper function for sc.pp.scale
        """
        sc.pp.scale(
            self.adata,
            zero_center=True,
            max_value=10,
            copy=False,
            layer=None,
            obsm=None
        )
        
    @utils.add_log
    def pca(self):
        """
        Wrapper function for sc.pp.pca
        """
        sc.pp.pca(
            self.adata,
            n_comps=50,
            zero_center=True,
            svd_solver='auto',
            random_state=0,
            return_info=False,
            use_highly_variable=True,
            dtype='float32',
            copy=False,
            chunked=False,
            chunk_size=None
        )       

    @utils.add_log
    def neighbors(self, ):
        """
        Wrapper function for sc.pp.neighbors(), for supporting multiple n_neighbors
        """
        sc.pp.neighbors(
            self.adata,
            n_neighbors=15,
            n_pcs=N_PCS,
            use_rep=None,
            knn=True,
            random_state=0,
            method='umap',
            metric='euclidean',
            key_added=None,
            copy=False
        )

    @utils.add_log
    def tsne(self):
        """
        Wrapper function for sc.tl.tsne, for supporting named slot of tsne embeddings
        """
        sc.tl.tsne(
            self.adata,
            n_pcs=N_PCS,
            n_jobs=self.thread,
            copy=False,
        )

    @utils.add_log
    def umap(self, ):
        """
        Wrapper function for sc.tl.umap, for supporting named slot of umap embeddings
        """
        sc.tl.umap(
            self.adata,
            min_dist=0.5,
            spread=1.0,
            n_components=2,
            maxiter=None,
            alpha=1.0,
            gamma=1.0,
            negative_sample_rate=5,
            init_pos='spectral',
            random_state=0,
            a=None,
            b=None,
            copy=False,
            method='umap',
            neighbors_key=None
        )

        
    @utils.add_log
    def leiden(self):
        """
        Wrapper function for sc.tl.leiden
        """
        sc.tl.leiden(
            self.adata,
            resolution=RESOLUTION,
            restrict_to=None,
            random_state=0,
            key_added='cluster',
            adjacency=None,
            directed=True,
            use_weights=True,
            n_iterations=-1,
            partition_type=None,
            neighbors_key=None,
            obsp=None,
            copy=False
        )

    @utils.add_log
    def find_marker_genes(self):
        """
        Wrapper function for sc.tl.rank_genes_groups
        """
        sc.tl.rank_genes_groups(
            self.adata,
            "cluster",
            reference='rest',
            pts=True,
            method="wilcoxon",
            use_raw=False,
            layer=NORMALIZED_LAYER
        )

    @utils.add_log
    def write_markers(self):
        '''
        write only p_val_adj < PVAL_CUTOFF to avoid too many markers
        '''
        df_markers = sc.get.rank_genes_groups_df(self.adata, group=None, pval_cutoff=PVAL_CUTOFF)
        df_markers = df_markers[df_markers['logfoldchanges'].notna()]
        markers_name_dict = {
            'group':'cluster',
            'names':'gene',
            'logfoldchanges':'avg_log2FC',
            'pvals':'p_val',
            'pvals_adj':'p_val_adj',
            'pct_nz_group':'pct.1',
            'pct_nz_reference':'pct.2'
         }
        df_markers = df_markers.rename(markers_name_dict,axis='columns')
        df_markers['cluster'] = df_markers['cluster'].map(lambda x : int(x)+1)
        df_markers = df_markers.loc[df_markers['p_val_adj'] < PVAL_CUTOFF, ]
        df_markers.to_csv(self.df_marker_raw_file, index=None, sep='\t')

        df_markers_filter = df_markers.loc[df_markers['avg_log2FC'] > 0].sort_values('p_val_adj').groupby('cluster').head(100)
        df_markers_filter = df_markers_filter.round({
            'avg_log2FC':3,
            'pct.1':3,
            'pct.2':3,
        })
        df_markers_filter.to_csv(self.df_marker_file, index=None, sep='\t')


    @utils.add_log
    def write_tsne(self):
        df_tsne = self.adata.obsm.to_df()[['X_tsne1','X_tsne2']]
        df_tsne['cluster']=self.adata.obs.cluster
        df_tsne['Gene_Counts']=self.adata.obs.n_genes_by_counts
        tsne_name_dict={'X_tsne1':'tSNE_1', 'X_tsne2':'tSNE_2'}
        df_tsne = df_tsne.rename(tsne_name_dict,axis='columns')
        df_tsne['cluster'] = df_tsne['cluster'].map(lambda x:int(x)+1)
        df_tsne.to_csv(self.df_tsne_file, sep='\t')

    @utils.add_log
    def write_h5ad(self):
        self.adata.write(self.h5ad_file)
    

    @utils.add_log
    def run(self):

        self.calculate_qc_metrics()
        self.write_mito_stats()
        self.normalize()
        self.hvg()
        self.scale()
        self.pca()
        self.neighbors()
        self.tsne()
        self.umap()
        self.leiden()
        self.find_marker_genes()

        self.write_markers()
        self.write_tsne()
        self.write_h5ad()


    def get_df(self):
        """
        return df_tsne, df_marker
        """
        df_tsne = read_tsne(self.df_tsne_file)
        df_marker = pd.read_csv(self.df_marker_file, sep="\t")
        df_marker = format_df_marker(df_marker)
        return df_tsne, df_marker


def get_opts_analysis_match(parser, sub_program):
    """
    Do not perform analysis. Only read data from scRNA-seq match_dir.
    """
    if sub_program:
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--tsne_file", help=HELP_DICT['tsne_file'])
        parser.add_argument("--df_marker_file", help=HELP_DICT['df_marker_file'])

        parser = s_common(parser)

class Report_runner(Step):

    def __init__(self, args, display_title=None):
    
        super().__init__(args, display_title=display_title)

    def add_marker_help(self):
        self.add_help_content(
            name='Marker Genes by Cluster',
            content='differential expression analysis based on the non-parameteric Wilcoxon rank sum test'
        )
        self.add_help_content(
            name='avg_log2FC',
            content='log fold-change of the average expression between the cluster and the rest of the sample'
        )
        self.add_help_content(
            name='pct.1',
            content='The percentage of cells where the gene is detected in the cluster'
        )
        self.add_help_content(
            name='pct.2',
            content='The percentage of cells where the gene is detected in the rest of the sample'
        )
        self.add_help_content(
            name='p_val_adj',
            content='Adjusted p-value, based on bonferroni correction using all genes in the dataset'
        )

    @staticmethod
    def get_df_file(match_dir):
        """
        return df_tsne_file, df_marker_file
        """
        match_dict = utils.parse_match_dir(match_dir)
        df_tsne_file = match_dict['tsne_coord']
        df_marker_file = match_dict.get('markers', None)
        return df_tsne_file, df_marker_file

    def get_df(self):
        """
        return df_tsne, df_marker
        """
        if utils.check_arg_not_none(self.args, 'match_dir'):
            df_tsne_file, df_marker_file = self.get_df_file(self.args.match_dir)
        elif utils.check_arg_not_none(self.args, 'tsne_file'):
            df_tsne_file = self.args.tsne_file
            df_marker_file = self.args.df_marker_file
        else:
            raise ValueError('match_dir or tsne_file must be specified')
        df_tsne = read_tsne(df_tsne_file)
        if df_marker_file:
            df_marker = pd.read_csv(df_marker_file, sep="\t")
            df_marker = format_df_marker(df_marker)
        else:
            df_marker = None
        return df_tsne, df_marker

    def run(self):
        pass



def pytest_addoption(parser):
    parser.addoption("--assays", action="store", default="")
    parser.addoption("--test_dir", help='test_dir', action="store", default="/SGRNJ03/randd/user/zhouyiqi/multi_tests/")


def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".
    assays_value = metafunc.config.option.assays
    test_dir_value = metafunc.config.option.test_dir
    if 'assays' in metafunc.fixturenames and assays_value is not None:
        metafunc.parametrize("assays", [assays_value])
    if 'test_dir' in metafunc.fixturenames and test_dir_value is not None:
        metafunc.parametrize("test_dir", [test_dir_value])



import unittest
import os
import cProfile

import celescope.snp.variant_calling as vc
from tests.unittests.__init__ import TEST_DIR_ROOT



class Test_snp(unittest.TestCase):
    def setUp(self):
        os.chdir(f'{TEST_DIR_ROOT}/snp/')
        self.sample = 'test1'
        self.vcf_file = f'{self.sample}/07.variant_calling/{self.sample}_merged.vcf'

    @unittest.skip('skip')
    def test_cell_UMI(self):
        cid = 7996
        outdir = f'{self.sample}/07.variant_calling/'
        df_UMI = vc.cell_UMI(cid, outdir, self.vcf_file)
        print(df_UMI)

    @unittest.skip('skip')
    def test_cell_UMI_2(self):
        cid = 2301
        os.chdir('/SGRNJ03/randd/RD20081701_SCOPEv2_Dynaseq/20210820/')
        sample = 'Mito_9geneMix_0812'
        outdir = f'{sample}/07.variant_calling/'
        vcf_file = f'{sample}/07.variant_calling/{sample}_merged.vcf'
        df_UMI = vc.cell_UMI(cid, outdir, vcf_file)
        print(df_UMI)


if __name__ == '__main__':
    unittest.main()


import subprocess
import sys

from celescope.tools import utils
from celescope.tools.make_ref import MakeRef_STAR


def parse_attributes(attrs):
    dic = {}
    for attr_str in attrs.split(';'):
        if attr_str:
            attr, val = attr_str.split('=')
            val = set(val.split(','))
            dic[attr] = val
    return dic

class Mkref_rna(MakeRef_STAR):
    def __init__(self, genome_type, args):
        super().__init__(genome_type, args)
        self.files['gtf'] = self.args.gtf
        self.files['mt_gene_list'] = self.args.mt_gene_list

    @utils.add_log
    def build_rna_star_index(self):
        SA = self._get_SA()
        cmd = (
            f'STAR \\\n'
            f'--runMode genomeGenerate \\\n'
            f'--runThreadN {self.args.thread} \\\n'
            f'--genomeDir ./ \\\n'
            f'--genomeFastaFiles {self.args.fasta} \\\n'
            f'--sjdbGTFfile {self.args.gtf} \\\n'
            f'--sjdbOverhang 100 \\\n'
            f'--genomeSAindexNbases {SA} \\\n'
        )

        sys.stderr.write(cmd+'\n')
        if not self.dry_run:
            subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        self.build_rna_star_index()


def mkref(args):
    genome_type = 'rna'
    with Mkref_rna(genome_type, args) as runner:
        runner.run()

def get_opts_mkref(parser, sub_program):
    MakeRef_STAR.opts(parser, sub_program)
    if sub_program:
        parser.add_argument(
            "--gtf",
            help="Required. Gtf file name.",
            required=True
        )
        parser.add_argument(
            "--mt_gene_list",
            help="""Mitochondria gene list file name. This file is a plain text file with one gene per line. 
If not provided, will use `MT-` and `mt-` to determine mitochondria genes.""",
            default="None"
        )
        parser.add_argument(
            '--attributes', 
            help='Attributes to keep. Example: `gene_biotype=protein_coding,lncRNA,antisense;`',
            default="gene_biotype=protein_coding,lncRNA,antisense,IG_LV_gene,IG_V_gene,IG_V_pseudogene,IG_D_gene,IG_J_gene,IG_J_pseudogene,IG_C_gene,IG_C_pseudogene,TR_V_gene,TR_V_pseudogene,TR_D_gene,TR_J_gene,TR_J_pseudogene,TR_C_gene;",
        )



import os,re
import sys
import pandas as pd
import anndata
import pysam
from multiprocessing import Pool
import subprocess

from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.dynaseq.__init__ import DYNA_MATRIX_DIR_SUFFIX
from celescope.rna.mkref import Mkref_rna
from celescope.tools.matrix import CountMatrix, Features, ROW, COLUMN
from celescope.tools import reference
from celescope.tools.plotly_plot import Tsne_plot, Violin_plot

toolsdir = os.path.dirname(__file__)


class Replacement(Step):
    """
    Features
    - Quantify unlabeled and labeled RNA.
    - Boxplots for TOR rates distribution.
    - TSNE plot for TOR rate 

    Output
    - `{sample}.labeled.h5ad` h5ad file contains ['total', 'labeled', 'unlabeled'] layers and TOR rate of each cell/gene.
    - `{sample}_labeled_feature_bc_matrix` The labeled expression matrix of cell barcodes & all features in Matrix Market Exchange Formats. (will be deprecated in future versions)
    - `{sample}_unlabeled_feature_bc_matrix` The unlabeled expression matrix of cell barcodes & all features in Matrix Market Exchange Formats. (will be deprecated in future versions)
    - `{sample}_labeled_detail.txt`  tab-delimited  file:
        - Barcode: Cell barcode sequence
        - UMI: UMI sequence
        - geneID: gene ID
        - TC: TC site number in a read (backgroup snp removed)
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # input files
        self.outdir = args.outdir
        self.sample = args.sample
        self.thread = int(args.thread)
        self.inbam = args.bam
        self.bcfile = args.cell
        self.snp_file = args.bg
        self.tsne = args.tsne
        self.cellsplit = args.cellsplit

        # set
        self.cell_dict, self.cell_num = utils.barcode_list_stamp(self.bcfile,cut=self.cellsplit)
        gtf_file = Mkref_rna.get_config(args.genomeDir)['files']['gtf']
        gp = reference.GtfParser(gtf_file)
        gp.get_id_name()
        self.features = gp.get_features()
        self.used_gene_id, self.used_gene_name = [], []
        self.used_features = None
        self.totaldf = pd.DataFrame()
        self.newdf, self.olddf = pd.DataFrame(), pd.DataFrame()
        self.adata = anndata.AnnData()
        self.cell_list = []
        self.bg = None

        # output files
        ## tmp outputdir
        self.tmp_dir = f'{args.outdir}/tmp'
        utils.check_mkdir(self.tmp_dir)
        ## final outputs
        self.h5ad = f'{self.out_prefix}.labeled.h5ad'
        self.detail_txt = f'{self.out_prefix}_labeled_detail.txt'
        self.dir_labeled = f'{self.outdir}/{DYNA_MATRIX_DIR_SUFFIX[0]}'
        self.dir_unlabeled = f'{self.outdir}/{DYNA_MATRIX_DIR_SUFFIX[1]}'
        self.outs = [self.dir_labeled, self.dir_unlabeled, self.h5ad]


    @utils.add_log
    def run(self):
        # get backgroud snp        
        self.bg = self.background_snp()
        # replacement
        self.run_quant()
        # stat and plot
        self.tor_plot()
        self.add_help()
        # output dedup and clean 
        self.clean_tmp()


    @utils.add_log
    def run_quant(self):
        ## set Parallelism para
        cell_arr = []
        fetch_arr = [self.inbam] * len(self.cell_dict)
        snp_list = [self.bg] * len(self.cell_dict)
        for x in self.cell_dict:
            cell_arr.append(self.cell_dict[x])
            self.cell_list.extend(self.cell_dict[x])
    
        mincpu = min(self.cell_num, self.thread)
        with Pool(mincpu) as pool:
            results = pool.starmap(Replacement.modify_bam, 
                      zip(fetch_arr,snp_list,cell_arr))
        # merge matrix
        for i in results:
            self.totaldf = pd.concat([self.totaldf,i])
        self.newdf = self.totaldf[self.totaldf['TC']>0]
        self.olddf = self.totaldf[self.totaldf['TC']==0]
        self.totaldf.to_csv(self.detail_txt, sep="\t", index=False)

        used_gene = self.totaldf['geneID'].unique()        
        for index,item in enumerate(self.features.gene_id):
            if item in used_gene:
                self.used_gene_id.append(item)
                self.used_gene_name.append(self.features.gene_name[index])
        self.used_features = Features(self.used_gene_id,self.used_gene_name)

        # output
        tmp_newdf = self.newdf.groupby([COLUMN, ROW]).agg({'UMI': 'count'})
        tmp_olddf = self.olddf.groupby([COLUMN, ROW]).agg({'UMI': 'count'})
        self.write_sparse_matrix(tmp_newdf, self.dir_labeled)
        self.write_sparse_matrix(tmp_olddf, self.dir_unlabeled)
        self.write_h5ad(self.totaldf, self.newdf, self.olddf)


    @staticmethod
    def modify_bam(bam, bg, cells):
        save = pysam.set_verbosity(0)
        bamfile = pysam.AlignmentFile(bam, 'rb')
        pysam.set_verbosity(save)
        readdict= {}

        for read in bamfile.fetch(until_eof=True):
            try:
                cb = read.get_tag('CB')
                if cb not in cells:
                    continue
                chro = read.reference_name
                ub = read.get_tag('UB')
                gene = read.get_tag('GX')
                tctag = 0
                true_tc = []

                if read.get_tag('ST') == '+':
                    stag = read.get_tag('TL')
                else:
                    stag = read.get_tag('AL')
                if len(stag) == 1 and stag[0] == 0:
                    tctag = 0
                    true_tc = stag
                else:
                    for si in range(0, len(stag)):
                        pos = chro + '_' + str(stag[si])
                        if pos not in bg:
                            true_tc.append(int(stag[si]))
                    tctag = len(true_tc)
                ## dedup: select the most TC read per UMI_gene
                readid = ':'.join([cb,ub,gene])
                if readid not in readdict:
                    readdict[readid] = [tctag,read,true_tc]
                else:
                    if tctag > readdict[readid][0]:
                        readdict[readid] = [tctag,read,true_tc]

            except (ValueError, KeyError):
                continue
        bamfile.close()

        ## count df
        tc_df = pd.DataFrame.from_dict(readdict,orient='index', columns=["TC","read","loc"])
        tc_df.reset_index(inplace=True)
        ub_df = tc_df['index'].str.split(':',expand=True)
        ub_df.columns = ['Barcode','UMI','geneID']
        #ub_df['Barcode'] = [cell] * ub_df.shape[0]
        outframe = pd.concat([ub_df,tc_df['TC']], axis=1)

        return outframe

    @staticmethod
    def createTag(d):
        return ''.join([''.join(key) + str(d[key]) + ';' for key in d.keys()])[:-1]

    @staticmethod
    def modifySCTag(sc,cnt,sstag):
        pattern = re.compile(rf'{sstag}\d+')
        result = pattern.sub(f'{sstag}{cnt}', sc)
        return result

    @utils.add_log
    def background_snp(self):
        outdict = {}
        bgs = []
        for bgargv in self.snp_file:
            if ',' in bgargv:
                bgs += bgargv.strip().split(',')
            else:
                bgs.append(bgargv)
        
        for bgfile in bgs:
            if bgfile.endswith('.csv'):
                df = pd.read_csv(bgfile, dtype={"chrom":str})
                if 'pos' in df.columns:
                    df['chrpos'] = df['chrom']+'_'+df['pos'].astype(str)
                else: #compatible with previous version
                    df['chrpos'] = df['chrom']+'_'+df['pos2'].astype(str)
                df1 = df[['chrpos','convs']]
                df1.set_index('chrpos',inplace=True)
                for key1 in df1.index.to_list():
                    outdict[key1] = 1
                #outdict.update(df1.to_dict(orient='index'))
            elif bgfile.endswith('.vcf'):
                bcf_in = pysam.VariantFile(bgfile)
                for rec in bcf_in.fetch():
                    try:
                        chrom, pos = rec.chrom, rec.pos
                        chr_pos = chrom+'_'+str(pos-1)
                        outdict[chr_pos] = 1
                    except (ValueError, KeyError):
                        continue
                bcf_in.close()
            else:
                try:
                    sys.exit(1)
                except SystemExit:
                    print('Background snp file format cannot be recognized! Only csv or vcf format.')
                finally:
                    print('Background snp file format cannot be recognized! Only csv or vcf format.')
        return outdict


    @utils.add_log
    def write_sparse_matrix(self, df, matrix_dir):
        count_matrix = CountMatrix.from_dataframe(df, self.features, barcodes=self.cell_list)
        count_matrix.to_matrix_dir(matrix_dir)

    @utils.add_log
    def write_h5ad(self, df, df_new, df_old):
        layers = {}
        matrix = CountMatrix.dataframe_to_matrix(df, self.used_features, barcodes=self.cell_list)
        layers['total'] = matrix
        layers['labeled'] = CountMatrix.dataframe_to_matrix(df_new, self.used_features, barcodes=self.cell_list)
        layers['unlabeled'] = CountMatrix.dataframe_to_matrix(df_old, self.used_features, barcodes=self.cell_list)
        
        adata = anndata.AnnData(
            X=matrix,
            obs=pd.DataFrame(index=pd.Series(self.cell_list, name='cell')),
            var=pd.DataFrame(index=pd.Series(self.used_gene_id, name='gene_id'), data={'gene_name': pd.Categorical(self.used_gene_name)}),
            layers=layers
        )
        adata = self.tor_stat(adata)
        adata.write(self.h5ad)
        self.adata = adata

    @utils.add_log
    def tor_stat(self, adata):
        cell_ntr = adata.layers['labeled'].sum(axis=1) / adata.layers['total'].sum(axis=1)
        gene_ntr = adata.layers['labeled'].sum(axis=0) / adata.layers['total'].sum(axis=0)
        adata.obs['TOR'] = cell_ntr
        adata.var['TOR'] = gene_ntr.T
        return adata

    @utils.add_log
    def tor_plot(self):
        tsne = pd.read_csv(self.tsne,sep="\t",index_col=0)
        tsne = tsne.loc[self.adata.obs.index]
        self.adata.obs['tSNE_1'] = tsne['tSNE_1']
        self.adata.obs['tSNE_2'] = tsne['tSNE_2']

        tsne_tor = Tsne_plot(self.adata.obs.sort_values(by="TOR"), 'TOR', discrete=False)
        tsne_tor.set_color_scale("PuRd")        
        self.add_data(tsne_tor=tsne_tor.get_plotly_div())

        vln_gene = Violin_plot(self.adata.var['TOR'], 'gene', color='#1f77b4').get_plotly_div()
        self.add_data(violin_gene=vln_gene)
        vln_cell = Violin_plot(self.adata.obs['TOR'], 'cell', color='#ff7f0e').get_plotly_div()
        self.add_data(violin_cell=vln_cell)

    @utils.add_log
    def add_help(self):
        self.add_help_content(
            name='TOR:',
            content='(RNA turn-over rate) Fraction of labeled transcripts per gene or cell.'
        )


    def run_cmd(self, cmd):
        subprocess.call(' '.join(cmd), shell=True)

    @utils.add_log
    def clean_tmp(self):
        cmd = (f'rm -rf {self.tmp_dir}')
        self.debug_subprocess_call(cmd)


@utils.add_log
def replacement(args):
    if args.control:
        return
    with Replacement(args, display_title='Labeled') as runner:
        runner.run()


def get_opts_replacement(parser, sub_program):
    parser.add_argument('--genomeDir',help=HELP_DICT['genomeDir'])
    parser.add_argument("--control", action='store_true',
                        help="For control samples to generate backgroup snp files and skip replacement")
    parser.add_argument("--cellsplit", default=300, type=int, help='split N cells into a list')
    if sub_program:
        parser.add_argument("--bam", 
            help='convsrion tagged bam from conversion step', required=True)
        parser.add_argument("--cell", help='barcode cell list', required=True)
        parser.add_argument('--bg', nargs='+', required=False,
                            help='background snp file, csv or vcf format')
        parser.add_argument('--tsne', help='tsne file from analysis step', required=True)
        parser = s_common(parser)
    return parser



import shutil
import os
import sys
import subprocess

import numpy as np
import pandas as pd

from celescope.tools.__init__ import FILTERED_MATRIX_DIR_SUFFIX, RAW_MATRIX_DIR_SUFFIX, COUNTS_FILE_NAME
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.tools.matrix import CountMatrix
from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.rna.mkref import Mkref_rna

class Cells_metrics(Step):
    @utils.add_log
    def add_cells_metrics(self, n_cells, fraction_reads_in_cells, mean_used_reads_per_cell, median_umi_per_cell,
        total_genes, median_genes_per_cell, saturation, valid_reads):
        self.add_metric(
            name='Estimated Number of Cells',
            value=n_cells,
            help_info='The number of barcodes considered as cell-associated.'
        )

        self.add_metric(
            name='Fraction Reads in Cells',
            value=fraction_reads_in_cells,
            value_type='fraction',
            help_info='the fraction of uniquely-mapped-to-transcriptome reads with cell-associated barcodes'
        )

        mean_reads_per_cell = valid_reads // n_cells
        self.add_metric(
            name='Mean Reads per Cell',
            value=mean_reads_per_cell,
            help_info='the number of Valid Reads divided by Estimated Number of Cells'
        )

        self.add_metric(
            name='Mean Used Reads per Cell',
            value=mean_used_reads_per_cell,
            help_info='the number of uniquely-mapped-to-transcriptome reads per cell-associated barcode'
        )

        self.add_metric(
            name='Median UMI per Cell',
            value=median_umi_per_cell,
            help_info='the median number of UMI counts per cell-associated barcode'
        )

        self.add_metric(
            name='Total Genes',
            value=total_genes,
            help_info='the number of genes with at least one UMI count in any cell'
        )

        self.add_metric(
            name='Median Genes per Cell',
            value=median_genes_per_cell,
            help_info='the median number of genes detected per cell-associated barcode'
        )

        self.add_metric(
            name='Saturation',
            value=saturation,
            value_type='fraction',
            help_info='the fraction of read originating from an already-observed UMI. '
        )

    def run(self):
        pass

class Cells(Cells_metrics):
    def __init__(self, args, display_title=None):
        os.chdir(args.root_dir)
        args.outdir = f'{args.sample}/cells'
        super().__init__(args, display_title=display_title)
        self.raw_matrix = f'{self.outs_dir}/{RAW_MATRIX_DIR_SUFFIX}'
        self.old_filtered_matrix = f'{self.outs_dir}/{FILTERED_MATRIX_DIR_SUFFIX}'
        self.counts_file = f'{self.outs_dir}/{COUNTS_FILE_NAME}'

        # out
        self.filter_matrix = f'{self.outdir}/{FILTERED_MATRIX_DIR_SUFFIX}'
        self.default_filter_matrix = f'{self.outs_dir}/default_{FILTERED_MATRIX_DIR_SUFFIX}'
        self.default_report_html = f'{self.outdir}/../default_{self.sample}_report.html'

        if os.path.exists(self.default_filter_matrix):
            self.old_filtered_matrix = self.default_filter_matrix

        self.outs = [self.filter_matrix]

    @utils.add_log
    def force_cells(self):
        raw = CountMatrix.from_matrix_dir(self.raw_matrix)
        df_counts = pd.read_csv(self.counts_file, index_col=0, header=0, sep='\t')
        bcs = list(df_counts.head(self.args.force_cells).index)
        filtered = raw.slice_matrix_bc(bcs)
        self.add_metric(
            name='Force cells', 
            value=self.args.force_cells,
            show=False,
        )
        return filtered

    @utils.add_log
    def filter_min_gene(self, filtered:CountMatrix) -> CountMatrix:
        bc_geneNum, _ = filtered.get_bc_geneNum()
        bc_indices = [bc for bc in bc_geneNum if bc_geneNum[bc] >= self.args.min_gene]
        n_filtered_cells = len(bc_geneNum) - len(bc_indices)
        self.add_metric(
            name='Minimum gene number in cells', 
            value=self.args.min_gene,
            show=False,
        )
        self.add_metric(
            name='Number of cells filtered by min_gene', 
            value=n_filtered_cells,
            show=False,
        )
        return filtered.slice_matrix(bc_indices)

    @utils.add_log
    def metrics_report(self, filtered:CountMatrix):
        p_default_dict = {
            self.old_filtered_matrix: self.default_filter_matrix,
            self.report_html: self.default_report_html,
        }
        for p, default_p in p_default_dict.items():
            if not os.path.exists(default_p):
                shutil.move(p, default_p)

        self.add_slot_step(
            slot='metrics',
            step_name='default_cells',
            val=self.old_step_dict['metrics'],
        )
        
        self.add_slot_step(
            slot='data',
            step_name='default_cells',
            val=self.old_step_dict['data'],
        )
        bcs = filtered.get_barcodes()
        n_cells = len(bcs)

        df_counts = pd.read_csv(self.counts_file, index_col=0, header=0, sep='\t')
        reads_total = df_counts['countedU'].sum()
        reads_cell = df_counts.loc[bcs, 'countedU'].sum()
        fraction_reads_in_cells = float(reads_cell / reads_total)
        mean_used_reads_per_cell = int(reads_cell // len(bcs))
        median_umi_per_cell = int(df_counts.loc[bcs, 'UMI'].median())

        bc_geneNum, total_genes = filtered.get_bc_geneNum()        
        median_genes_per_cell = int(np.median(list(bc_geneNum.values())))

        saturation = self.old_step_dict['metrics']['Saturation'] / 100
        df_counts.loc[:, 'mark'] = 'UB'
        df_counts.loc[bcs, 'mark'] = 'CB'
        df_counts.to_csv(self.counts_file, sep='\t', index=True)
        valid_reads = self.old_step_dict['metrics']['Mean Reads per Cell'] * self.old_step_dict['metrics']['Estimated Number of Cells']
        self.add_cells_metrics(n_cells, fraction_reads_in_cells, mean_used_reads_per_cell, median_umi_per_cell, total_genes, median_genes_per_cell, saturation, valid_reads)
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.counts_file))

    @utils.add_log
    def filter_mito(self, filtered:CountMatrix) -> CountMatrix:
        if not self.args.genomeDir:
            raise ValueError("--genomeDir must be provided when --max_mito is set.")
        mt_gene_list = Mkref_rna.get_config(self.args.genomeDir)['files']['mt_gene_list']
        if not mt_gene_list:
            raise FileNotFoundError(f'mt_gene_list not exist in genomeDir: {self.args.genomeDir}')
        mito_genes, _ = utils.read_one_col(mt_gene_list)
        bc_fraction = filtered.get_genes_fraction(mito_genes)
        bc_indices = (bc_fraction <= self.args.max_mito).nonzero()[1]
        n_filtered_cells = len(filtered.get_barcodes()) - len(bc_indices)
        filtered = filtered.slice_matrix(bc_indices)
        self.add_metric(
            name='Maximum fraction of mitocondrial UMI in cells', 
            value=self.args.max_mito,
            show=False,
        )
        self.add_metric(
            name='Number of cells filtered by max_mito', 
            value=n_filtered_cells,
            show=False,
        )
        return filtered

    @utils.add_log
    def soloCellFilter(self, filtered:CountMatrix) -> CountMatrix:
        temp_raw = f'{self.outdir}/raw' 
        cmd = f'cp -r {self.raw_matrix} {self.outdir}; gunzip {temp_raw}/*.gz' #STAR do not recognize gzip matrix
        if not os.path.exists(temp_raw):
            subprocess.check_call(cmd, shell=True)
        temp_filtered = f'./{self.outdir}/filtered/'
        cmd = f'STAR --runMode soloCellFiltering {temp_raw} {temp_filtered} --soloCellFilter {self.args.soloCellFilter}'
        subprocess.check_call(cmd, shell=True)
        cmd = f'gzip {temp_filtered}/*'
        subprocess.check_call(cmd, shell=True)
        filtered = CountMatrix.from_matrix_dir(temp_filtered)
        self.add_metric(
            name='soloCellFilter', 
            value=self.args.soloCellFilter,
            show=False,
        )
        return filtered

    @utils.add_log
    def run(self):
        if self.args.max_mito > 1.0:
            sys.exit('max_mito should be less than 1.0')
        filtered = CountMatrix.from_matrix_dir(self.old_filtered_matrix)
        if self.args.force_cells > 0:
            filtered= self.force_cells()
        elif self.args.soloCellFilter:
            filtered = self.soloCellFilter(filtered)
        if self.args.max_mito < 1.0:
            filtered = self.filter_mito(filtered)
        if self.args.min_gene > 0:
            filtered = self.filter_min_gene(filtered)

        filtered.to_matrix_dir(self.filter_matrix)
        self.metrics_report(filtered)

def cells(args):
    with Cells(args) as runner: 
        runner.run()


def get_opts_cells(parser, sub_program=True):
    if sub_program:
        parser.add_argument(
            '--force_cells',
            help='Force to use this number of cells.',
            default=0,
            type=int,
        )
        parser.add_argument(
            '--soloCellFilter',
            help='The same as the argument in STARsolo. Ignored when --force_cells is set.',
            default='',
        )
        parser.add_argument(
            '--root_dir',
            help='The root directory of CeleScope runs.',
            default='./',
        )
        parser.add_argument(
            '--max_mito',
            help='Maximum mitocondrial fraction in a cell.',
            default='1.0',
            type=float,
        )
        parser.add_argument(
            '--min_gene',
            help='Minimum gene number in a cell.',
            default=0,
            type=int,
        )
        parser.add_argument(
            '--genomeDir',
            help=HELP_DICT['genomeDir'],           
        )
        parser = s_common(parser)


#!/usr/bin/env python3

import sys
import argparse
import math

isotypeRanks = {"IGHM":0, "IGHD":1, "IGHG3":2, "IGHG1":3, "IGHA1":4, "IGHG2":5, "IGHG4":6, "IGHE":7, "IGHA2":8, "*":9, ".":9}
isotypeOrder = ["IGHM", "IGHD", "IGHG3", "IGHG1", "IGHA1", "IGHG2", "IGHG4", "IGHE", "IGHA2"]
chainOrder = ["IGH", "IGK", "IGL", "TRA", "TRB", "TRG", "TRD"]

def GetChainType(v, j, c):
	s = ""
	if (c != "*" and c != "."):
		s = c 
	elif (j != "*" and j != "."):
		s = j
	elif (v != "*" and v != "."):
		s = v
	else:
		return -1
	
	if (s[0:3] == "IGH"):
		return (0, isotypeRanks[c])
	elif (s[0:3] == "IGK"):
		return 1
	elif (s[0:3] == "IGL"):
		return 2
	elif (s[0:3] == "TRA"):
		return 3
	elif (s[0:3] == "TRB"):
		return 4
	elif (s[0:3] == "TRG"):
		return 5
	elif (s[0:3] == "TRD"):
		return 6
	else:
		return -1

def ComputeRichness(rep):
	return len(rep)

def ComputeCPK(rep):
	if (len(rep) == 0):
		return "NA"
	return len(rep) / sum(list(rep.values())) * 1000

def ComputeEntropy(rep):
	if (len(rep) == 0):
		return "NA"
	total = sum(list(rep.values()))
	return sum([-x/total*math.log(x/total) for x in rep.values()]) 

def ComputeClonality(rep):
	if (len(rep) <= 1):
		return "NA"
	return 1 - ComputeEntropy(rep) / math.log(len(rep))

def OutputChain(rep, name):
	outputList = [name]
	outputList.append(sum(list(rep.values())))
	outputList.append(ComputeRichness(rep))
	outputList.append(ComputeCPK(rep))
	outputList.append(ComputeEntropy(rep))
	outputList.append(ComputeClonality(rep))
	print("\t".join([str(x) for x in outputList]))

def ProcessImmuneRepertoire(immrep):
	print("\t".join(["#chain", "Abundance", "Richness", "CPK", "Entropy", "Clonality"]))
	# IGH overall
	tmp = {}
	for i in range(10):
		for c in immrep[(0,i)]:
			if (c not in tmp):
				tmp[c] = 0
			tmp[c] += immrep[(0,i)][c]
	OutputChain(tmp, "IGH")
	# IGH chains
	for i in range(9):
		OutputChain(immrep[(0, i)], isotypeOrder[i])
	for i in range(1, 7):
		OutputChain(immrep[i], chainOrder[i])

if (__name__ == "__main__"):
	parser = argparse.ArgumentParser(description="Immune repertoire diversity statistics")
	parser.add_argument("-r", help="repertoire file", dest="repfile", required=True)
	parser.add_argument("-f", help="repertoire file format (TRUST4_report, TRUST4_barcode_report,...)", dest="format", default="TRUST4_report")
	parser.add_argument("--ntaa", help="Use nucleotide(nt) or amino acids(aa)", dest="ntaa", default="aa")
	args = parser.parse_args()

	immrep = {}
	for i in range(10):
		immrep[(0, i)] = {}
	for i in range(1, 7):
		immrep[i] = {}

	fp = open(args.repfile)
	if (args.format == "TRUST4_report"):
		for line in fp:
			if (line[0] == "#" or line[0:5] == "count"):
				continue
			cols = line.rstrip().split()
			chainType = GetChainType(cols[4], cols[6], cols[7])
			if ("_" in cols[3] or cols[3] == "partial" or "?" in cols[3] or 
				chainType == -1):
				continue
			if (cols[3] not in immrep[chainType]):
				immrep[chainType][cols[3]] = 0
			immrep[chainType][cols[3]] += int(cols[0])
	elif (args.format == "TRUST4_barcode_report"):
		for line in fp:
			if (line[0] == "#" or line[0:5] == "count"):
				continue
			mainCols = line.rstrip().split()
			for i in [2, 3]:
				if (mainCols[i] == "*"):
					continue
				cols = mainCols[i].split(",")
				chainType = GetChainType(cols[0], cols[2], cols[3])
				if ("_" in cols[5] or cols[5] == "partial" or "?" in cols[5] or
						chainType == -1):
					continue
				if (cols[5] not in immrep[chainType]):
					immrep[chainType][cols[5]] = 0
				immrep[chainType][cols[5]] += 1
	else:
		print("Unknown format " + args.format)
		sys.exit(1)
	fp.close()

	ProcessImmuneRepertoire(immrep) 



from celescope.rna.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_rna(Multi):
    """
    ## Usage
    ```
        multi_rna\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --thread 8\\
        --mod shell
    ```
    Work for both single cell RNA-Seq and single nuclei RNA-Seq.
    """

    def starsolo(self, sample):
        step = 'starsolo'
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def analysis(self, sample):
        step = 'analysis'
        matrix_file = f'{self.outdir_dic[sample]["outs"]}/filtered'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--matrix_file {matrix_file} '
        )
        self.process_cmd(cmd, step, sample, m=10, x=1)

def main():
    multi = Multi_rna(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



STEPS = ["mkref", "sample", "barcode", "cutadapt", "consensus", "mapping_vdj", "count_vdj"]
__ASSAY__ = 'vdj'

CHAINS = {
    "TCR": ["TRA", "TRB"],
    "BCR": ["IGH", "IGL", "IGK"],
}

PAIRS = {
    "TCR": [["TRA", "TRB"]],
    "BCR": [["IGH", "IGK"], ["IGH", "IGL"]],
}


import io
import json
import os

import pandas as pd
from jinja2 import Environment, FileSystemLoader, select_autoescape

env = Environment(
    loader=FileSystemLoader(os.path.dirname(__file__) + '/../templates/'),
    autoescape=select_autoescape(['html', 'xml']),
)


class reporter:

    def __init__(
        self, assay, name, outdir, sample,
        stat_file=None, plot=None, parameters=None,
        table_file=None, table_header=None, df=None, html_flag=True
    ):
        self.assay = assay
        self.name = name
        self.stat_file = stat_file
        self.outdir = outdir
        self.sample = sample
        self.plot = plot
        self.parameters = parameters
        self.table_file = table_file
        self.table_header = table_header
        self.df = df
        self.html_flag = html_flag

    def get_report(self):

        json_file = self.outdir + '/.data.json'
        if not os.path.exists(json_file):
            data = {}
        else:
            fh = open(json_file)
            data = json.load(fh)
            fh.close()

        if self.stat_file:
            df = pd.read_table(self.stat_file, header=None, sep=':', dtype=str)
            data[self.name + '_summary'] = df.values.tolist()

        if self.plot:
            data[self.name + '_plot'] = self.plot

        if self.parameters:
            data[self.name + '_parameters'] = self.parameters

        if self.table_file:
            df = pd.read_csv(self.table_file, sep="\t")
            df = df.fillna(value="")
            data[self.name + '_table'] = df.values.tolist()

        if isinstance(self.df, pd.DataFrame):
            df = self.df.fillna(value="")
            data[self.name + '_table'] = df.values.tolist()

        if self.table_header:
            data[self.name + '_table_header'] = self.table_header

        if self.html_flag:
            template = env.get_template(f'html/{self.assay}/base.html')
            report_html = "{outdir}/{sample}_report.html".format(
                outdir=self.outdir, sample=self.sample)
            with io.open(report_html, 'w', encoding='utf8') as fh:
                html = template.render(data)
                # fh.write(html.encode('utf-8'))
                fh.write(html)

        with open(json_file, 'w') as fh:
            json.dump(data, fh, indent=4)



import json

import pysam
import numpy as np

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT, HELP_INFO_DICT


def get_opts_count_bam(parser, sub_program):
    parser.add_argument("--min_query_length", help='Minimum query length.', default=35)

    if sub_program:
        parser.add_argument('--match_dir', help=HELP_DICT['match_dir'], required=True)
        parser.add_argument('--capture_bam', required=True)
        s_common(parser)


class Count_bam(Step):

    def __init__(self, args, display_title='Count'):
        super().__init__(args, display_title)

        # set
        self.min_query_length = int(args.min_query_length)
        self.capture_bam = args.capture_bam

        # read barcodes
        match_dir_dict = utils.parse_match_dir(args.match_dir)
        self.match_barcode = match_dir_dict['match_barcode']
        self.n_match_barcode = match_dir_dict['n_match_barcode']
        self.add_metric(
            name=HELP_INFO_DICT['matched_barcode_number']['display'],
            value=self.n_match_barcode,
            help_info=HELP_INFO_DICT['matched_barcode_number']['info']
        )

        # data
        self.total_corrected_umi = 0
        self.count_dict = utils.genDict(dim=3)

        # out
        self.raw_read_count_file = f'{self.out_prefix}_raw_read_count.json'

    @utils.add_log
    def process_bam(self):
        # process bam
        samfile = pysam.AlignmentFile(self.args.capture_bam, "rb")
        for read in samfile:
            ref = read.reference_name
            query_length = read.infer_query_length()
            attr = read.query_name.split(':')
            barcode = attr[0]
            umi = attr[1]
            if (barcode in self.match_barcode) and (query_length >= self.min_query_length):
                self.count_dict[barcode][ref][umi] += 1

    @utils.add_log
    def add_some_metrics(self):
        read_count_list = []
        for barcode in self.count_dict:
            for ref in self.count_dict[barcode]:
                for umi in self.count_dict[barcode][ref]:
                    read_count_list.append(self.count_dict[barcode][ref][umi])
        mean_read_count_per_umi = round(float(np.mean(read_count_list)), 2)
        self.add_metric(
            name='Mead Read Count per UMI',
            value=mean_read_count_per_umi,
            help_info='you can use this value to determine `min_support_read`'
        )

        n_positive_cell = len(self.count_dict)
        self.add_metric(
            name='Number of positive cells',
            value=n_positive_cell,
            total=self.n_match_barcode,
            help_info='this number is calculated on raw read counts(i.e. before filtering)'
        )

    @utils.add_log
    def write_count_file(self):

        with open(self.raw_read_count_file, 'w') as fp:
            json.dump(self.count_dict, fp, indent=4)

    @utils.add_log
    def run(self):
        self.process_bam()
        self.add_some_metrics()
        self.write_count_file()





import re
import collections
import csv
import sys

import pandas as pd

from celescope.tools import utils
from celescope.tools.matrix import Features

PATTERN = re.compile(r'(\S+?)\s*"(.*?)"')
gtf_row = collections.namedtuple( 'gtf_row', 'seqname source feature start end ' \
                                 'score strand frame attributes' )

def row2list(row):
    '''
    row: gtf_row
    '''
    attr_str = '; '.join( [ '%s "%s"' % attr for attr in row.attributes.items() ] )
    return [
        row.seqname,
        row.source,
        row.feature,
        row.start,
        row.end,
        row.score,
        row.strand,
        row.frame,
        attr_str
    ]

class GtfParser:
    def __init__(self, gtf_fn):
        self.gtf_fn = gtf_fn
        self.gene_id = []
        self.gene_name = []
        self.id_name = {}
        self.id_strand = {}

    def get_properties_dict(self, properties_str):
        """
        allow no space after semicolon
        """
        
        if isinstance(properties_str, dict):
            return properties_str

        properties = collections.OrderedDict()
        attrs = properties_str.split(';')
        for attr in attrs:
            if attr:
                m = re.search(PATTERN, attr)
                if m:
                    key = m.group(1).strip()
                    value = m.group(2).strip()
                    properties[key] = value

        return properties

    def gtf_reader_iter(self):
        """
        Yield:
            row: list
            gtf_row
        """
        with utils.generic_open(self.gtf_fn, mode='rt') as f:
            reader = csv.reader(f, delimiter='\t')
            for i, row in enumerate(reader, start=1):
                if len(row) == 0:
                    continue
                if row[0].startswith('#'):
                    yield row, None
                    continue

                if len(row) != 9:
                    sys.exit(f"Invalid number of columns in GTF line {i}: {row}\n")

                if row[6] not in ['+', '-']:
                    sys.exit(f"Invalid strand in GTF line {i}: {row}\n")
                
                seqname = row[0]
                source  = row[1]
                feature = row[2]
                # gff/gtf is 1-based, end-inclusive
                start   = int(row[3])
                end     = int(row[4])
                score   = row[5]
                strand  = row[6]
                frame  = row[7]
                attributes = self.get_properties_dict(row[8])

                yield row, gtf_row( seqname, source, feature, \
                    start, end, score, strand, frame, \
                    attributes )

    @utils.add_log
    def get_id_name(self):
        """
        return: {gene_id:gene_name}
        """
        for _, grow in self.gtf_reader_iter():
            if not grow:
                continue
            gene_id = grow.attributes['gene_id']
            self.id_strand[gene_id] = grow.strand
            if 'gene_name' not in grow.attributes:
                gene_name = gene_id
            else:
                gene_name = grow.attributes['gene_name']

            if gene_id not in self.id_name:
                self.gene_id.append(gene_id)
                self.gene_name.append(gene_name)
                self.id_name[gene_id] = gene_name

        return self.id_name
                    
    def get_features(self):
        """
        Returns:
            Features object
        """
        if not self.gene_id:
            sys.exit("Empty self.gene_id. Run self.get_id_name first.")
        features = Features(self.gene_id, self.gene_name)
        return features

    def get_strand(self):
        return self.id_strand

class GtfBuilder:
    def __init__(self, in_gtf_fn, out_gtf_fn, attributes, add_intron=False):
        self.in_gtf_fn = in_gtf_fn
        self.out_gtf_fn = out_gtf_fn
        self.attributes = attributes
        self.add_intron = add_intron
        self.mt_gene_list = 'mt_gene_list.txt'
    
    @staticmethod
    def get_introns(exons):
        sys.stderr.write( 'done (%d exons).\n' % len(exons) )
        # add intron
        transcripts = collections.defaultdict(list)
        for grow in exons:
            if 'transcript_id' in grow.attributes:
                transcripts[grow.attributes['transcript_id']].append(grow)
        sys.stderr.write( 'done (%d transcripts).\n' % len(transcripts) )
        
        introns = []
        for (transcript_id,rows) in transcripts.items():
            if len(set( (row.seqname,row.strand) for row in rows )) != 1:
                sys.stderr.write( 'Malformed transcript "%s". Skipping.' % transcript_id)

            rows.sort( key = lambda row: row.start )
            starts = [ row.start for row in rows ]
            ends   = [ row.end   for row in rows ]

            for (i,j) in zip(ends[:-1],starts[1:]):
                assert i < j
                if i+1 <= j-1:
                    intron_row = gtf_row(
                            seqname = rows[0].seqname,
                            source  = rows[0].source,
                            feature = 'intron',
                            start   = i+1,
                            end     = j-1,
                            score   = '.',
                            strand  = rows[0].strand,
                            frame   = '.',
                            attributes = rows[0].attributes )

                    introns.append( intron_row )
        sys.stderr.write( 'done (%d introns).\n' % len(introns) )
        return introns


    @utils.add_log
    def build_gtf(self):
        """
        add intron
        Filter gene biotypes
        """
        sys.stderr.write("Writing GTF file...\n")
        gp = GtfParser(self.in_gtf_fn)
        n_filter = 0
        exons = []
        mt = set()

        with open(self.out_gtf_fn, 'w') as f:
            writer = csv.writer(f, delimiter='\t', quoting=csv.QUOTE_NONE, quotechar='')
            for row, grow in gp.gtf_reader_iter():
                if not grow:
                    writer.writerow(row)
                    continue            

                remove = False
                for key, value in grow.attributes.items():
                    if key in self.attributes and value not in self.attributes[key]:
                        remove = True

                if not remove:
                    if grow.seqname.upper() == 'MT':
                        mt.add(grow.attributes['gene_name'])
                    writer.writerow(row)
                    if grow.feature == 'exon':
                        exons.append(grow)
                else:
                    n_filter += 1
            sys.stderr.write(f'filtered line number: {n_filter}\n')

            if self.add_intron:                
                introns = self.get_introns(exons)
                for intron in introns:
                    writer.writerow(row2list(intron))

            if mt:
                sys.stderr.write('Writing mito genes\n')
                pd.Series(list(mt)).to_csv(self.mt_gene_list, index=False, header=False)

    





        




import random
from collections import namedtuple

import numpy as np
import numpy.ma as ma

import celescope.tools.emptydrop_cr.sgt as cr_sgt  # # modified sgt.py
import celescope.tools.emptydrop_cr.stats as cr_stats  # # modified stats.py
from celescope.tools.matrix import CountMatrix
from celescope.tools import utils


# Set random seed
random.seed(0)
np.random.seed(0)

# Number of additional barcodes to consider after the initial cell calling
N_CANDIDATE_BARCODES = 20000

# Number of partitions (max number of barcodes to consider for ambient estimation)
N_PARTITIONS = 90000

# Drop this top fraction of the barcodes when estimating ambient.
MAX_OCCUPIED_PARTITIONS_FRAC = 0.5

# Minimum number of UMIS per barcode to consider after the initial cell calling
MIN_UMIS = 500

# Minimum ratio of UMIs to the median (initial cell call UMI) to consider after the initial cell calling
MIN_UMI_FRAC_OF_MEDIAN = 0.01

# Maximum adjusted p-value to call a barcode as non-ambient
MAX_ADJ_PVALUE = 0.01


def adjust_pvalue_bh(p):
    """ Multiple testing correction of p-values using the Benjamini-Hochberg procedure """
    descending = np.argsort(p)[::-1]
    # q = p * N / k where p = p-value, N = # tests, k = p-value rank
    scale = float(len(p)) / np.arange(len(p), 0, -1)
    q = np.minimum(1, np.minimum.accumulate(scale * p[descending]))

    # Return to original order
    return q[np.argsort(descending)]


def estimate_profile_sgt(matrix, barcode_indices, nz_feat):
    """ Estimate a gene expression profile by Simple Good Turing.
    Args:
      raw_mat (sparse matrix): Sparse matrix of all counts
      barcode_indices (np.array(int)): Barcode indices to use
      nz_feat (np.array(int)): Indices of features that are non-zero at least once
    Returns:
      profile (np.array(float)): Estimated probabilities of length len(nz_feat).
    """
    # Initial profile estimate
    prof_mat = matrix[:, barcode_indices]

    profile = np.ravel(prof_mat[nz_feat, :].sum(axis=1))
    zero_feat = np.flatnonzero(profile == 0)

    # Simple Good Turing estimate
    p_smoothed, p0 = cr_sgt.sgt_proportions(profile[np.flatnonzero(profile)])

    # Distribute p0 equally among the zero elements.
    p0_i = p0/len(zero_feat)

    profile_p = np.repeat(p0_i, len(nz_feat))
    profile_p[np.flatnonzero(profile)] = p_smoothed

    assert np.isclose(profile_p.sum(), 1.0)
    return profile_p


# Construct a background expression profile from barcodes with <= T UMIs
def est_background_profile_sgt(matrix, use_bcs):
    """ Estimate a gene expression profile on a given subset of barcodes.
         Use Good-Turing to smooth the estimated profile.
    Args:
      matrix (scipy.sparse.csc_matrix): Sparse matrix of all counts
      use_bcs (np.array(int)): Indices of barcodes to use (col indices into matrix)
    Returns:
      profile (use_features, np.array(float)): Estimated probabilities of length use_features.
    """
    # Use features that are nonzero anywhere in the data
    use_feats = np.flatnonzero(np.asarray(matrix.sum(1)))

    # Estimate background profile
    bg_profile_p = estimate_profile_sgt(matrix, use_bcs, use_feats)

    return (use_feats, bg_profile_p)

@utils.add_log
def find_nonambient_barcodes(raw_mat, recovered_cells,
                             min_umi_frac_of_median=MIN_UMI_FRAC_OF_MEDIAN,
                             min_umis_nonambient=MIN_UMIS,
                             max_adj_pvalue=MAX_ADJ_PVALUE,):
    """ Call barcodes as being sufficiently distinct from the ambient profile
    Args:
      raw_mat: raw matrix of UMI counts
      recovered_cells: expected number of recovered cells
    Returns:
    TBD
    """
    NonAmbientBarcodeResult = namedtuple('NonAmbientBarcodeResult',
                                         ['eval_bcs',      # Candidate barcode indices (n)
                                          'log_likelihood',  # Ambient log likelihoods (n)
                                          'pvalues',       # pvalues (n)
                                          'pvalues_adj',   # B-H adjusted pvalues (n)
                                          'is_nonambient',  # Boolean nonambient calls (n)
                                          ])

    # Estimate an ambient RNA profile
    umis_per_bc = np.squeeze(np.asarray(raw_mat.sum(axis=0)))
    # get the index of sorted umis_per_bc (ascending, bc_order[0] is the index of the smallest element in umis_per_bc)
    bc_order = np.argsort(umis_per_bc)

    # Require non-zero barcodes
    nz_bcs = np.flatnonzero(umis_per_bc)

    # Take what we expect to be the barcodes associated w/ empty partitions.
    # Drop the top MAX_OCCUPIED_PARTITIONS_FRAC fraction of barcodes
    empty_bcs = bc_order[::-1][int(N_PARTITIONS*MAX_OCCUPIED_PARTITIONS_FRAC):N_PARTITIONS]
    empty_bcs.sort()

    # Get the non-zero barcodes overlapped with barcodes assumed to be associated w/ empty partitions
    use_bcs = np.intersect1d(empty_bcs, nz_bcs, assume_unique=True)

    if len(use_bcs) > 0:
        try:
            # Get used "Gene" features (eval_features)
            # and the smoothed prob profile per "Gene" (ambient_profile_p)
            eval_features, ambient_profile_p = est_background_profile_sgt(raw_mat.tocsc(), use_bcs)
        except cr_sgt.SimpleGoodTuringError as e:
            print(str(e))
    else:
        eval_features = np.zeros(0, dtype=int)
        ambient_profile_p = np.zeros(0)

    # Choose candidate cell barcodes
    # Regular ordmag filter
    gg_filtered_indices, gg_filtered_metrics, _msg = cr_stats.filter_cellular_barcodes_ordmag(
        umis_per_bc, recovered_cells=recovered_cells)


    orig_cell_bc_set = set(gg_filtered_indices)
    orig_cells = np.flatnonzero(np.fromiter((bc in orig_cell_bc_set for bc in range(raw_mat.shape[1])), dtype=bool))

    # No good incoming cell calls
    if orig_cells.sum() == 0:
        print('Error: No original cells are selected!')
        return None, None, None

    # Look at non-cell barcodes above a minimum UMI count
    eval_bcs = np.ma.array(np.arange(raw_mat.shape[1]))
    eval_bcs[orig_cells] = ma.masked

    median_initial_umis = np.median(umis_per_bc[orig_cells])

    min_umis = int(max(min_umis_nonambient, round(np.ceil(median_initial_umis * min_umi_frac_of_median))))

    eval_bcs[umis_per_bc < min_umis] = ma.masked
    n_unmasked_bcs = len(eval_bcs) - eval_bcs.mask.sum()

    # Take the top N_CANDIDATE_BARCODES by UMI count, of barcodes that pass the above criteria
    # For evaluation of non-ambient bcs using background info estimated from SGT
    eval_bcs = np.argsort(ma.masked_array(umis_per_bc, mask=eval_bcs.mask))[:n_unmasked_bcs][-N_CANDIDATE_BARCODES:]

    if len(eval_bcs) == 0:
        find_nonambient_barcodes.logger.warning('no eval bcs are selected to evaluate non-empty bcs from SGT results!')
        return orig_cells, gg_filtered_metrics, None
    else:
        assert not np.any(np.isin(eval_bcs, orig_cells))

        eval_mat = raw_mat.tocsc()[eval_features, :][:, eval_bcs]

        if len(ambient_profile_p) == 0:
            return orig_cells, gg_filtered_metrics, None

        # Compute observed log-likelihood of barcodes being generated from ambient RNA
        obs_loglk = cr_stats.eval_multinomial_loglikelihoods(eval_mat, ambient_profile_p)

        # Simulate log likelihoods
        distinct_ns, sim_loglk = cr_stats.simulate_multinomial_loglikelihoods(ambient_profile_p, umis_per_bc[eval_bcs],
                                                                              num_sims=10000, verbose=True)

        # Compute p-values
        pvalues = cr_stats.compute_ambient_pvalues(umis_per_bc[eval_bcs], obs_loglk, distinct_ns, sim_loglk)

        pvalues_adj = adjust_pvalue_bh(pvalues)
        max_adj_pvalue = 0.01
        is_nonambient = pvalues_adj <= max_adj_pvalue

        filtered_bc_indices = np.concatenate((orig_cells, eval_bcs[is_nonambient]), axis=None)

        return filtered_bc_indices, gg_filtered_metrics, NonAmbientBarcodeResult(
            eval_bcs=eval_bcs,
            log_likelihood=obs_loglk,
            pvalues=pvalues,
            pvalues_adj=pvalues_adj,
            is_nonambient=is_nonambient,
        )


def cell_calling_3(all_matrix_dir, expected_cell_num):

    count_matrix = CountMatrix.from_matrix_dir(matrix_dir=all_matrix_dir)

    # Run cell calling
    filtered_bc_indices, round_1_filtered_metrics, _non_ambient_barcode_result = find_nonambient_barcodes(
        raw_mat=count_matrix.get_matrix(), recovered_cells=expected_cell_num)

    raw_barcodes = np.array(count_matrix.get_barcodes())
    cell_bc = raw_barcodes[filtered_bc_indices]
    initial_cell_num = round_1_filtered_metrics['filtered_bcs']
    return cell_bc, initial_cell_num




import numpy as np
import pysam
import sys
import subprocess

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT
from celescope.snp.__init__ import PANEL


def get_gene_list(args):
    if args.panel:
        gene_list = utils.get_gene_region_from_bed(args.panel)[0]
        n_gene = len(gene_list)
    else:
        gene_list, n_gene = utils.read_one_col(args.gene_list)
    
    return set(gene_list), n_gene
        



class Target_metrics(Step):
    """
    ## Features
    - Filter bam file
        - Filter reads that are not cell-associated.
        - Filter reads that are not mapped to target genes. 

    - Collect enrichment metrics.

    ## Output
    - `filtered.bam` BAM file after filtering. Reads that are not cell-associated or not mapped to target genes are filtered.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # set
        self.match_barcode_list, self.n_cell = utils.get_barcode_from_match_dir(args.match_dir)
        self.match_barcode = set(self.match_barcode_list)

        self.gene_list, self.n_gene = get_gene_list(args)

        if not self.gene_list:
            sys.exit("You must provide either --panel or --gene_list!")

        self.count_dict = utils.genDict(dim=3, valType=int)

        self.add_metric(
            name="Number of Target Genes",
            value=self.n_gene,
        )
        self.add_metric(
            name="Number of Cells",
            value=self.n_cell,
        )

        # out file
        self.out_bam_file = f'{self.out_prefix}_filtered.bam'
        self.out_bam_file_sorted = f'{self.out_prefix}_filtered_sorted.bam'

    @utils.add_log
    def read_bam_write_filtered(self):
        sam_temp = f'{self.out_bam_file}.temp'
        with pysam.AlignmentFile(self.args.bam, "rb") as reader:
            header = reader.header.to_dict()
            # add RG to header
            if self.args.add_RG:
                header['RG'] = []
                for barcode in self.match_barcode_list:
                    header['RG'].append({
                        'ID': barcode,
                        'SM': barcode,
                    })
            with pysam.AlignmentFile(sam_temp, "w", header=header) as writer:
                for record in reader:
                    try:
                        gene_name = record.get_tag('GN')
                    except KeyError:
                        continue
                    # compatible with tag bam
                    try:
                        barcode = record.get_tag('CB')
                        UMI = record.get_tag('UB')
                    except KeyError:
                        continue
                    if barcode in self.match_barcode and gene_name in self.gene_list:
                        if self.args.add_RG:
                            record.set_tag(tag='RG', value=record.get_tag('CB'), value_type='Z')
                        writer.write(record)
                    self.count_dict[barcode][gene_name][UMI] += 1

            cmd = f'samtools view -b {sam_temp} -o {self.out_bam_file}; rm {sam_temp}'
            subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def parse_count_dict_add_metrics(self):
        total_reads = 0
        enriched_reads = 0
        enriched_reads_in_cells = 0
        enriched_reads_per_cell_list = []

        for barcode in self.count_dict:
            cell_enriched_read = 0
            for gene_name in self.count_dict[barcode]:
                gene_read = sum(self.count_dict[barcode][gene_name].values())
                total_reads += gene_read
                if gene_name in self.gene_list:
                    enriched_reads += gene_read
                    if barcode in self.match_barcode:
                        enriched_reads_in_cells += gene_read
                        cell_enriched_read += gene_read
            if barcode in self.match_barcode:
                enriched_reads_per_cell_list.append(cell_enriched_read)

        self.parse_count_dict_add_metrics.logger.debug(
            f'enriched_reads_per_cell_list: '
            f'{sorted(enriched_reads_per_cell_list)}'
            f'len: {len(enriched_reads_per_cell_list)}'
        )

        valid_enriched_reads_per_cell_list = [cell for cell in enriched_reads_per_cell_list if cell > 0]
        n_valid_cell = len(valid_enriched_reads_per_cell_list)
        self.add_metric(
            name="Number of Valid Cells",
            value=n_valid_cell,
            total=self.n_cell,
            help_info='Cells with at least one read mapped to the target genes.'
        )
        self.add_metric(
            name="Enriched Reads",
            value=enriched_reads,
            total=total_reads,
        )
        self.add_metric(
            name="Enriched Reads in Cells",
            value=enriched_reads_in_cells,
            total=total_reads,
        )
        self.add_metric(
            name="Median Enriched Reads per Valid Cell",
            value=np.median(valid_enriched_reads_per_cell_list),
        )

    def run(self):
        self.read_bam_write_filtered()
        self.parse_count_dict_add_metrics()
        samtools_runner = utils.Samtools(
            self.out_bam_file,
            self.out_bam_file_sorted,
            self.args.thread,
            debug=self.debug,
        )
        samtools_runner.sort_bam()
        samtools_runner.index_bam()


@utils.add_log
def target_metrics(args):
    with Target_metrics(args, display_title='Target Enrichment') as runner:
        runner.run()


def get_opts_target_metrics(parser, sub_program):
    parser.add_argument("--gene_list", help=HELP_DICT['gene_list'])
    parser.add_argument("--panel", help=HELP_DICT['panel'], choices=list(PANEL))
    if sub_program:
        parser.add_argument("--bam", help='Input bam file', required=True)
        parser.add_argument('--match_dir', help=HELP_DICT['match_dir'], required=True)
        parser.add_argument(
            '--add_RG', help='Add tag read group: RG. RG is the same as CB(cell barcode)', action='store_true')
        parser = s_common(parser)
